import abc
import logging
import warnings
from collections import defaultdict
from typing import Any, Literal, Union

import networkx
import openff.fragmenter
from openff.fragmenter.chemi import (
    assign_elf10_am1_bond_orders,
    extract_fragment,
    find_ring_systems,
    find_stereocenters,
)
from openff.fragmenter.states import _enumerate_stereoisomers
from openff.fragmenter.utils import (
    default_functional_groups,
    get_atom_index,
    get_map_index,
    global_toolkit_registry,
)
from openff.toolkit.topology import Atom, Molecule
from openff.toolkit.utils import (
    GLOBAL_TOOLKIT_REGISTRY,
    ToolkitRegistry,
    ToolkitWrapper,
)
from openff.toolkit.utils.exceptions import AtomMappingWarning

try:
    from pydantic.v1 import BaseModel, Field
except ImportError:
    from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)

BondTuple = tuple[int, int]
AtomAndBondSet = tuple[set[int], set[BondTuple]]

Stereochemistries = dict[Union[int, BondTuple], str]
RingSystems = dict[int, AtomAndBondSet]

FunctionalGroups = dict[str, AtomAndBondSet]

Heuristic = Literal["path_length", "wbo"]


class Fragment(BaseModel):
    """An object which stores minimal information about a molecules fragment."""

    smiles: str = Field(
        ...,
        description="A mapped SMILES pattern describing the fragment. The map indices "
        "assigned to each atom in the pattern will correspond to the map index of the "
        "corresponding parent atom. If an atom does not have a map index it is likely "
        "that the atom was added (either H, or C) to ensure every atom in the fragment "
        "has a sensible valence.",
    )

    bond_indices: tuple[int, int] = Field(
        ...,
        description="The map indices of the atoms involved in the bond that the "
        "fragment was built around.",
    )

    @property
    def molecule(self) -> Molecule:
        """The fragment represented as an OpenFF molecule object."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AtomMappingWarning)

            return Molecule.from_smiles(self.smiles, allow_undefined_stereo=True)


class FragmentationResult(BaseModel):
    """An object which stores the results of fragmenting a molecule."""

    parent_smiles: str = Field(
        ...,
        description="A mapped SMILES pattern describing the parent molecule that was "
        "fragmented.",
    )

    fragments: list[Fragment] = Field(..., description="The generated fragments.")

    provenance: dict[str, Any] = Field(
        ...,
        description="A dictionary storing provenance information about how the "
        "fragments were generated.",
    )

    @property
    def parent_molecule(self) -> Molecule:
        """The parent molecule represented as an OpenFF molecule object."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AtomMappingWarning)

            return Molecule.from_smiles(self.parent_smiles)

    @property
    def fragment_molecules(self) -> dict[BondTuple, Molecule]:
        """A dictionary of the fragment molecules represented as OpenFF molecule
        objects, indexed by the map indices of the bond that each fragment was built
        around."""
        return {fragment.bond_indices: fragment.molecule for fragment in self.fragments}

    @property
    def fragments_by_bond(self) -> dict[BondTuple, Fragment]:
        """Returns a dictionary of fragments indexed by the bond (defined in terms of
        the map indices of the atoms that form it) that the fragment was built around.
        """
        return {fragment.bond_indices: fragment for fragment in self.fragments}


class Fragmenter(BaseModel, abc.ABC):
    """The base class that all fragmentation engines should inherit from."""

    functional_groups: dict[str, str] = Field(
        default_factory=default_functional_groups,
        description="A dictionary of SMARTS of functional groups that should not be "
        "fragmented, indexed by an informative name, e.g. 'alcohol': '[#6]-[#8X2H1]'.",
    )

    @classmethod
    def _find_stereo(cls, molecule: Molecule) -> Stereochemistries:
        """Find chiral atoms and bonds, store the chirality.

        Notes
        -----
            * This is needed to check if a fragment has flipped chirality. Currently
              this can happen and it is a bug.

        Parameters
        ----------
        molecule
            The molecule to search for stereochemistry.

        Returns
        -------
            The stereochemistry associated with atom and bond stereocenters
        """

        atom_stereo = {
            get_map_index(molecule, atom.molecule_atom_index): atom.stereochemistry
            for atom in molecule.atoms
            if atom.stereochemistry is not None
        }

        bond_stereo = {
            (
                get_map_index(molecule, bond.atom1_index),
                get_map_index(molecule, bond.atom2_index),
            ): bond.stereochemistry
            for bond in molecule.bonds
            if bond.stereochemistry is not None
        }

        return {**atom_stereo, **bond_stereo}

    @classmethod
    def _check_stereo(
        cls, fragment: Molecule, parent_stereo: Stereochemistries
    ) -> bool:
        """Checks if the stereochemistry of a fragment is different to the
        stereochemistry of the parent.

        Parameters
        ----------
        fragment
            The fragment whose stereo should be compared to the parent.
        parent_stereo
            The stereochemistry of the parent molecule.

        Returns
        -------
            Whether the fragment has the same stereochemistry as the parent.
        """

        atom_stereocenters, bond_stereocenters = find_stereocenters(fragment)

        # Check for new / flipped chiral centers.
        for atom_index in atom_stereocenters:
            map_index = get_map_index(fragment, atom_index)

            if map_index not in parent_stereo:
                logger.warning(f"A new stereocenter formed at atom {map_index}")
                return False

            fragment_stereo = fragment.atoms[atom_index].stereochemistry

            if fragment_stereo != parent_stereo[map_index]:
                logger.warning(
                    f"Stereochemistry for atom {map_index} flipped from "
                    f"{parent_stereo[map_index]} to {fragment_stereo}"
                )

                return False

        for index_tuple in bond_stereocenters:
            map_tuple = tuple(get_map_index(fragment, i) for i in index_tuple)

            map_tuple = (
                map_tuple if map_tuple in parent_stereo else tuple(reversed(map_tuple))
            )

            if map_tuple not in parent_stereo:
                logger.warning(f"A new chiral bond formed at bond {map_tuple}")
                return False

            fragment_stereo = fragment.get_bond_between(*index_tuple).stereochemistry

            if fragment_stereo != parent_stereo[map_tuple]:
                logger.warning(
                    f"Stereochemistry for bond {map_tuple} flipped from "
                    f"{parent_stereo[map_tuple]} to {fragment_stereo}"
                )

                return False

        return True

    @classmethod
    def _fix_stereo(
        cls, fragment: Molecule, parent_stereo: Stereochemistries
    ) -> Molecule | None:
        """Flip all stereocenters and find the stereoisomer that matches the parent

        Parameters
        ----------
        fragment
            The fragment whose stereochemistry should be flipped to match the parent.
        parent_stereo
            The stereochemistry of the parent molecule.

        Returns
        -------
            A fragment with the same stereochemistry as parent molecule if possible else ``None``.
        """

        for stereoisomer in _enumerate_stereoisomers(fragment, 200, True):
            if not cls._check_stereo(stereoisomer, parent_stereo):
                continue

            return stereoisomer

        return None

    @classmethod
    def _find_functional_groups(
        cls, molecule: Molecule, functional_groups: dict[str, str]
    ) -> FunctionalGroups:
        """Find the atoms and bonds involved in the functional groups specified by
        ``functional_groups``.

        Parameters
        ----------
        molecule
            The molecule to search for function groups.
        functional_groups
            A dictionary of SMARTS of functional groups that should not be fragmented
            indexed by a friendly string label, e.g. 'alcohol: [#6:1]-[#8H1X2:2]'

        Returns
        -------
            The atoms and bonds in the found function groups stored in a dictionary
            indexed by a unique key associated with each functional group.
        """

        found_groups = {}

        for functional_group, smarts in functional_groups.items():
            unique_matches = {
                tuple(sorted(match))
                for match in molecule.chemical_environment_matches(smarts)
            }

            for i, match in enumerate(unique_matches):
                atoms = {get_map_index(molecule, index) for index in match}
                bonds = {
                    (
                        get_map_index(molecule, bond.atom1_index),
                        get_map_index(molecule, bond.atom2_index),
                    )
                    for bond in molecule.bonds
                    if bond.atom1_index in match and bond.atom2_index in match
                }

                found_groups[f"{functional_group}_{i}"] = (atoms, bonds)

        return found_groups

    @classmethod
    def find_rotatable_bonds(
        cls, molecule: Molecule, target_bond_smarts: list[str] | None
    ) -> list[BondTuple]:
        """Finds the rotatable bonds in a molecule *including* rotatable double
        bonds.

        Parameters
        ----------
        molecule
            The molecule to search for rotatable bonds.
        target_bond_smarts
            An optional list of SMARTS patterns that should be used to identify the bonds
            within the parent molecule to grow fragments around. Each SMARTS pattern
            should include **two** indexed atoms that correspond to the two atoms
            involved in the central bond.

            If no pattern is provided fragments will be constructed around all 'rotatable
            bonds'. A 'rotatable bond' here means any bond matched by a
            `[!$(*#*)&!D1:1]-,=;!@[!$(*#*)&!D1:2]` SMARTS pattern with the added
            constraint that the **heavy** degree (i.e. the degree excluding hydrogen) of
            both atoms in the bond must be >= 2.

        Returns
        -------
            A list of the **map** indices of the atoms that form the rotatable
            bonds, ``[(m1, m2),...]``.
        """

        if target_bond_smarts is None:
            matches = molecule.chemical_environment_matches(
                "[!$(*#*)&!D1:1]-,=;!@[!$(*#*)&!D1:2]"
            )

        else:
            matches = [
                match
                for smarts in target_bond_smarts
                for match in molecule.chemical_environment_matches(smarts)
            ]

            if not all(len(match) == 2 for match in matches):
                raise ValueError(
                    f"The `target_bond_smarts` pattern ({target_bond_smarts}) "
                    f"must define exactly two indexed atoms to match."
                )

        unique_matches = {tuple(sorted(match)) for match in matches}

        if target_bond_smarts is None:
            # Drop bonds without a heavy degree of at least 2 on each end to avoid
            # finding terminal bonds
            def heavy_degree(atom_index: int) -> int:
                atom = molecule.atoms[atom_index]
                return sum(1 for atom in atom.bonded_atoms if atom.atomic_number != 1)

            unique_matches = {
                match
                for match in unique_matches
                if all(heavy_degree(i) > 1 for i in match)
            }

        return [
            (
                get_map_index(molecule, match[0]),
                get_map_index(molecule, match[1]),
            )
            for match in unique_matches
        ]

    @classmethod
    def _atom_bond_set_to_mol(
        cls,
        parent: Molecule,
        parent_stereo: Stereochemistries,
        atoms: set[int],
        bonds: set[BondTuple],
    ) -> tuple[Molecule, bool]:
        """Extracts a subset of a molecule based on a set of atom and bond indices.

        Parameters
        ----------
        parent
            The parent molecule to slice.
        parent_stereo
            The stereochemistry of the parent. This will be used to ensure the returned
            subset molecule retains the correct stereochemistry.
        atoms
            set of map indices
        bonds
            set of bond tuples (m1, m2)

        Returns
        -------
            The subset molecule and a flag to indicate whether any new stereocenters are present in the fragment.
        """

        fragment = extract_fragment(parent, atoms, bonds)

        if not cls._check_stereo(fragment, parent_stereo):
            fixed_fragment = cls._fix_stereo(fragment, parent_stereo)
            if fixed_fragment is None:
                return fragment, True
            else:
                return fixed_fragment, False

        return fragment, False

    @classmethod
    def _get_torsion_quartet(
        cls, molecule: Molecule, bond: BondTuple
    ) -> AtomAndBondSet:
        """Get all atoms bonded to the torsion quartet around rotatable bond

        Parameters
        ----------
        molecule
            The molecule containing the rotatable bond.
        bond
            map indices of atoms in bond

        Returns
        -------
            The map indices of atoms in quartet and the bonds in quartet.
        """

        atom_map_indices = {*bond}
        bond_map_indices = {bond}

        atoms = [
            molecule.atoms[i]
            for i, j in molecule.properties["atom_map"].items()
            if j in bond
        ]

        for atom in atoms:
            map_index = get_map_index(molecule, atom.molecule_atom_index)

            for neighbor in atom.bonded_atoms:
                neighbour_map_index = get_map_index(
                    molecule, neighbor.molecule_atom_index
                )

                atom_map_indices.add(neighbour_map_index)
                bond_map_indices.add((map_index, neighbour_map_index))

                for next_neighbour in neighbor.bonded_atoms:
                    next_neighbour_map_index = get_map_index(
                        molecule, next_neighbour.molecule_atom_index
                    )

                    atom_map_indices.add(next_neighbour_map_index)
                    bond_map_indices.add(
                        (neighbour_map_index, next_neighbour_map_index)
                    )

        return atom_map_indices, bond_map_indices

    @classmethod
    def _find_ring_systems(
        cls,
        molecule: Molecule,
        functional_groups: FunctionalGroups,
        keep_non_rotor_ring_substituents: bool = False,
    ) -> RingSystems:
        """This function finds all ring systems in a molecule.

        Parameters
        ----------
        molecule
            The molecule to search for ring systems.
        functional_groups
            A dictionary of the functional groups on the molecule which should not
            be fragmented.
        keep_non_rotor_ring_substituents
            If True, keep all non rotatable ring substituents. According to the
            benchmark, it is not necessary.

        Returns
        -------
            Any found ring systems.
        """

        atom_to_ring_indices = find_ring_systems(molecule)

        # Find the map indices of the atoms involved in each ring system.
        ring_system_atoms = {
            ring_index: {
                get_map_index(molecule, i)
                for i in atom_to_ring_indices
                if atom_to_ring_indices[i] == ring_index
            }
            for ring_index in {*atom_to_ring_indices.values()}
        }

        # Find the map indices of the bonds involved in each ring system.
        ring_system_bonds = defaultdict(set)

        for bond in molecule.bonds:
            ring_index_1 = atom_to_ring_indices.get(bond.atom1_index, -1)
            ring_index_2 = atom_to_ring_indices.get(bond.atom2_index, -2)

            if ring_index_1 != ring_index_2:
                continue

            ring_system_bonds[ring_index_1].add(
                (
                    get_map_index(molecule, bond.atom1_index),
                    get_map_index(molecule, bond.atom2_index),
                )
            )

        # Scan the neighbours of the ring system atoms for any functional groups
        # / non-rotor substituents which should be included in the ring systems.
        for ring_index in ring_system_atoms:
            # If any atoms are part of a functional group, include the other atoms in the
            # group in the ring system lists
            ring_functional_groups = {
                functional_group
                for map_index in ring_system_atoms[ring_index]
                for functional_group in functional_groups
                if map_index in functional_groups[functional_group][0]
            }

            ring_system_atoms[ring_index].update(
                map_index
                for functional_group in ring_functional_groups
                for map_index in functional_groups[functional_group][0]
            )
            ring_system_bonds[ring_index].update(
                map_tuple
                for functional_group in ring_functional_groups
                for map_tuple in functional_groups[functional_group][1]
            )

            if not keep_non_rotor_ring_substituents:
                continue

            non_rotor_atoms, non_rotor_bonds = cls._find_non_rotor_ring_substituents(
                molecule, ring_system_atoms[ring_index]
            )

            ring_system_atoms[ring_index].update(non_rotor_atoms)
            ring_system_bonds[ring_index].update(non_rotor_bonds)

        ring_systems = {
            ring_index: (
                ring_system_atoms[ring_index],
                ring_system_bonds[ring_index],
            )
            for ring_index in ring_system_atoms
        }

        return ring_systems

    @classmethod
    def _find_non_rotor_ring_substituents(
        cls, molecule: Molecule, ring_system_atoms: set[int]
    ) -> AtomAndBondSet:
        """Find the non-rotor substituents attached to a particular ring system.

        Parameters
        ----------
        molecule
            The molecule to search for non-rotor ring substituents.
        ring_system_atoms
            The map indices of the atoms in the ring system of interest.

        Returns
        -------
            The map indices of the atoms and bonds involved in any found
            functional groups.
        """

        rotatable_bonds = molecule.find_rotatable_bonds()

        def heavy_degree(atom: Atom) -> int:
            return sum(1 for atom in atom.bonded_atoms if atom.atomic_number != 1)

        rotor_bonds = [
            bond
            for bond in rotatable_bonds
            if heavy_degree(bond.atom1) >= 2 and heavy_degree(bond.atom2) >= 2
        ]

        non_rotor_atoms = set()
        non_rotor_bonds = set()

        for bond in molecule.bonds:
            # Check if the bond is a rotor.
            if bond in rotor_bonds:
                continue

            if bond.atom1.atomic_number == 1 or bond.atom2.atomic_number == 1:
                continue

            map_index_1 = get_map_index(molecule, bond.atom1_index)
            map_index_2 = get_map_index(molecule, bond.atom2_index)

            in_system_1 = map_index_1 in ring_system_atoms
            in_system_2 = map_index_2 in ring_system_atoms

            if (in_system_1 and in_system_2) or (not in_system_1 and not in_system_2):
                continue

            non_rotor_atoms.update((map_index_1, map_index_2))
            non_rotor_bonds.add((map_index_1, map_index_2))

        return non_rotor_atoms, non_rotor_bonds

    @classmethod
    def _get_ring_and_fgroups(
        cls,
        parent: Molecule,
        parent_groups: FunctionalGroups,
        parent_rings: RingSystems,
        atoms: set[int],
        bonds: set[BondTuple],
    ) -> AtomAndBondSet:
        """Adds the atom and bond indices of groups ortho to those already in
        a fragment, such that they are retained during fragmentation.

        Parameters
        ----------
        parent
            The molecule being fragmented.
        parent_groups
            A dictionary of the functional groups on the molecule which should not
            be fragmented.
        parent_rings
            A dictionary of the ring systems in the molecule which should not
            be fragmented.
        atoms
            The map indices of the atoms in the fragment.
        bonds
            The map indices of the bonds in the fragment.

        Returns
        -------
            The updated set of atom and bond map indices to retain.
        """

        # Find the sets of atoms which are located ortho to one of the bonds being
        # fragmented.
        ortho_atoms, ortho_bonds = cls._find_ortho_substituents(parent, bonds)

        atoms.update(ortho_atoms)
        bonds.update(ortho_bonds)

        # Include the rings systems and functional groups connected to the current
        # atom sets.
        new_atoms = set()
        new_bonds = set()

        fragment_groups = {
            group
            for group in parent_groups
            if any(atom in parent_groups[group][0] for atom in atoms)
        }

        for functional_group in fragment_groups:
            new_atoms.update(parent_groups[functional_group][0])
            new_bonds.update(parent_groups[functional_group][1])

        fragment_rings = {
            ring_index
            for ring_index in parent_rings
            if any(atom in parent_rings[ring_index][0] for atom in atoms)
        }

        for ring_system in fragment_rings:
            new_atoms.update(parent_rings[ring_system][0])
            new_bonds.update(parent_rings[ring_system][1])

        atoms.update(new_atoms)
        bonds.update(new_bonds)

        # Ensure the matched bonds doesn't include duplicates.
        bonds = {tuple(sorted(bond)) for bond in bonds}

        return atoms, bonds

    @classmethod
    def _find_ortho_substituents(
        cls, parent: Molecule, bonds: set[BondTuple]
    ) -> AtomAndBondSet:
        """Find ring substituents that are ortho to one of the rotatable bonds specified
        in a list of bonds.

        Parameters
        ----------
        parent
            The parent molecule being fragmented.
        bonds
            The map indices of the rotatable bonds.

        Returns
        -------
            The set of map indices of atoms in ortho group and of bond tuples in ortho
            group.
        """

        matched_atoms = set()
        matched_bonds = set()

        for match in parent.chemical_environment_matches(
            "[!#1:1]~&!@[*:2]@[*:3]~&!@[!#1*:4]"
        ):
            map_tuple = tuple(get_map_index(parent, i) for i in match)

            if map_tuple[:2] not in bonds and map_tuple[:2][::-1] not in bonds:
                continue

            matched_atoms.update(map_tuple[::3])
            matched_bonds.update((map_tuple[i], map_tuple[i + 1]) for i in [0, 2])

        # Ensure the matched bonds doesn't include duplicates.
        matched_bonds = {tuple(sorted(bond)) for bond in matched_bonds}

        return matched_atoms, matched_bonds

    @classmethod
    def _cap_open_valence(
        cls,
        parent: Molecule,
        parent_groups: FunctionalGroups,
        atoms: set[int],
        bonds: set[BondTuple],
    ) -> AtomAndBondSet:
        """Cap with methyl for fragments that ends with N, O or S. Otherwise cap with H

        Parameters
        ----------
        parent
            The molecule being fragmented.
        parent_groups
            A dictionary of the functional groups on the molecule which should not
            be fragmented.
        atoms
            The map indices of the atoms in the fragment being constructed.
        bonds
            The map indices of the bonds in the fragment being constructed.
        """

        map_index_to_functional_group = {
            map_index: functional_group
            for functional_group in parent_groups
            for map_index in parent_groups[functional_group][0]
        }

        atoms_to_add = set()
        bonds_to_add = set()

        for map_index in atoms:
            atom_index = get_atom_index(parent, map_index)
            atom = parent.atoms[atom_index]

            if (
                atom.atomic_number not in (7, 8, 16)
                and map_index not in map_index_to_functional_group
            ):
                continue

            # If atom is N, O or S, it needs to be capped
            should_cap = False

            for neighbour in atom.bonded_atoms:
                neighbour_map_index = get_map_index(
                    parent, neighbour.molecule_atom_index
                )

                if neighbour.atomic_number == 1 or neighbour_map_index in atoms:
                    continue

                should_cap = True
                break

            if not should_cap:
                continue

            for neighbour in atom.bonded_atoms:
                if neighbour.atomic_number != 6:
                    continue

                neighbour_map_index = get_map_index(
                    parent, neighbour.molecule_atom_index
                )

                atoms_to_add.add(neighbour_map_index)
                bonds_to_add.add((map_index, neighbour_map_index))

        atoms.update(atoms_to_add)
        bonds.update(bonds_to_add)

        return atoms, bonds

    @classmethod
    def _prepare_molecule(
        cls,
        molecule: Molecule,
        functional_groups: dict[str, str],
        keep_non_rotor_ring_substituents: bool,
    ) -> tuple[Molecule, Stereochemistries, FunctionalGroups, RingSystems]:
        """Prepare a molecule for fragmentation.

        This involves canonically ordering the molecule, determining the stereochemistry
        of any stereocenters, detecting any functional groups which should be preserved,
        and finding any ring systems which should be preserved.

        Parameters
        ----------
        molecule
            The parent molecule that should be fragmented.
        functional_groups:
            A dictionary of SMARTS of functional groups that should not be fragmented.
        keep_non_rotor_ring_substituents:
            If True, will always keep all non rotor substituents on ring.

        Returns
        -------
            The prepared molecule to fragment, its stereochemistry, and the functional
            groups and ring systems it contains that should not be fragmented.
        """

        # Canonically order the molecule to try and make the fragmentation more
        # deterministic.
        molecule: Molecule = molecule.canonical_order_atoms()
        molecule.properties["atom_map"] = {i: i + 1 for i in range(molecule.n_atoms)}

        # Keep track of stereo to make sure it does not flip
        stereo = cls._find_stereo(molecule)

        # Find the functional groups and ring systems which should not be fragmented.
        found_functional_groups = cls._find_functional_groups(
            molecule, functional_groups
        )
        found_ring_systems = cls._find_ring_systems(
            molecule, found_functional_groups, keep_non_rotor_ring_substituents
        )

        return molecule, stereo, found_functional_groups, found_ring_systems

    @abc.abstractmethod
    def _fragment(
        self,
        molecule: Molecule,
        target_bond_smarts: list[str] | None,
    ) -> FragmentationResult:
        """The internal implementation of ``fragment``.

        Parameters
        ----------
        molecule
            The molecule to fragment.
        target_bond_smarts
            An optional SMARTS pattern that should be used to identify the bonds within
            the parent molecule to grow fragments around.

        Returns
        -------
            The results of the fragmentation including the fragments and provenance
            about the fragmentation.
        """

        raise NotImplementedError()

    def fragment(
        self,
        molecule: Molecule,
        target_bond_smarts: list[str] | None = None,
        toolkit_registry: ToolkitRegistry | ToolkitWrapper | None = None,
    ) -> FragmentationResult:
        """Fragments a molecule according to this class' settings.

        Notes
        -----
        * This method is currently *not* guaranteed to be thread safe as it uses and
          modifies the OpenFF toolkits' ``GLOBAL_TOOLKIT_REGISTRY``.

        Parameters
        ----------
        molecule
            The molecule to fragment.
        target_bond_smarts
            An optional list of SMARTS patterns that should be used to identify the bonds
            within the parent molecule to grow fragments around. Each SMARTS pattern
            should include **two** indexed atoms that correspond to the two atoms
            involved in the central bond.

            If no pattern is provided fragments will be constructed around all 'rotatable
            bonds'. A 'rotatable bond' here means any bond matched by a
            `[!$(*#*)&!D1:1]-,=;!@[!$(*#*)&!D1:2]` SMARTS pattern with the added
            constraint that the **heavy** degree (i.e. the degree excluding hydrogen) of
            both atoms in the bond must be >= 2. Note this will not find terminal
            rotatable bonds such as -OH, -NH2 -CH3.
        toolkit_registry
            The underlying cheminformatics toolkits to use for things like conformer
            generation, WBO computation etc. If no value is provided, the current
            ``GLOBAL_TOOLKIT_REGISTRY`` will be used. See the OpenFF toolkit
            documentation for more information.

        Returns
        -------
            The results of the fragmentation including the fragments and provenance
            about the fragmentation.
        """

        if toolkit_registry is None:
            toolkit_registry = GLOBAL_TOOLKIT_REGISTRY

        with global_toolkit_registry(toolkit_registry):
            result = self._fragment(molecule, target_bond_smarts)

            result.provenance["toolkits"] = [
                (toolkit.__class__.__name__, toolkit.toolkit_version)
                for toolkit in GLOBAL_TOOLKIT_REGISTRY.registered_toolkits
            ]

        if "options" not in result.provenance:
            result.provenance["options"] = {}

        if target_bond_smarts is not None:
            result.provenance["options"]["target_bond_smarts"] = target_bond_smarts

        return result

    def _default_provenance(self) -> dict[str, Any]:
        """Returns a dictionary containing default provenance information."""

        provenance = {
            "creator": openff.fragmenter.__package__,
            "version": openff.fragmenter.__version__,
            "options": self.dict(),
        }

        return provenance


class WBOOptions(BaseModel):
    """A set of options for controlling how Wiberg Bond Orders are computed."""

    method: Literal["am1-wiberg-elf10"] = Field(
        "am1-wiberg-elf10", description="The method to use when computing the WBOs."
    )

    max_conformers: int = Field(
        800, description="The maximum number of conformers to average the WBOs over."
    )
    rms_threshold: float = Field(
        1.0,
        description="The minimum RMS value [Angstrom] at which two conformers are "
        "considered redundant and one is deleted.",
    )


class WBOFragmenter(Fragmenter):
    """Fragment engine for fragmenting molecules using Wiberg Bond Order."""

    scheme: Literal["WBO"] = "WBO"

    wbo_options: WBOOptions = Field(
        WBOOptions(), description="The options to use when computing the WBOs."
    )

    threshold: float = Field(
        0.03,
        description="The threshold for the central bond WBO. If the fragment WBO is "
        "below this threshold, fragmenter will grow out the fragment one bond at a "
        "time via the path specified by the heuristic option",
    )
    heuristic: Heuristic = Field(
        "path_length",
        description="The path fragmenter should take when fragment needs to be grown "
        "out. The options are ``['wbo', 'path_length']``.",
    )

    keep_non_rotor_ring_substituents: bool = Field(
        False,
        description="Whether to always keep all non rotor substituents on rings. If "
        "``False``, rotor substituents on rings will only be retained if they are "
        "ortho to the central bond or if it's needed for WBO to be within the "
        "threshold.",
    )

    def _fragment(
        self, molecule: Molecule, target_bond_smarts: list[str] | None
    ) -> FragmentationResult:
        """Fragments a molecule in such a way that the WBO of the bond that a fragment
        is being built around does not change beyond the specified threshold.
        """

        (
            molecule,
            stereochemistry,
            functional_groups,
            ring_systems,
        ) = self._prepare_molecule(
            molecule, self.functional_groups, self.keep_non_rotor_ring_substituents
        )

        # Calculate WBO for molecule
        if self.wbo_options.method != "am1-wiberg-elf10":
            raise NotImplementedError(
                "WBOs can currently only be computed using 'am1-wiberg-elf10'."
            )

        molecule = assign_elf10_am1_bond_orders(
            molecule, self.wbo_options.max_conformers, self.wbo_options.rms_threshold
        )

        rotatable_bonds = self.find_rotatable_bonds(molecule, target_bond_smarts)
        wbo_rotor_bonds = self._get_rotor_wbo(molecule, rotatable_bonds)

        fragments = {
            bond: self._build_fragment(
                molecule,
                stereochemistry,
                functional_groups,
                ring_systems,
                bond,
                wbo_rotor_bonds[bond],
                threshold=self.threshold,
                heuristic=self.heuristic,
            )
            for bond in wbo_rotor_bonds
        }

        return FragmentationResult(
            parent_smiles=molecule.to_smiles(mapped=True),
            fragments=[
                Fragment(smiles=fragment.to_smiles(mapped=True), bond_indices=bond)
                for bond, fragment in fragments.items()
            ],
            provenance=self._default_provenance(),
        )

    @classmethod
    def _get_rotor_wbo(
        cls, molecule: Molecule, rotor_bonds: list[BondTuple]
    ) -> dict[BondTuple, float]:
        """Cache the WBO of each bond in a specific set of rotor bonds..

        Parameters
        ----------
            molecule
                The molecule containing the rotors.
            rotor_bonds
                The map indices of the rotor bonds to return the WBOs of.

        Returns
        -------
            The WBO of each rotor bond.
        """

        if any(bond.fractional_bond_order is None for bond in molecule.bonds):
            raise RuntimeError(
                "WBO was not calculated for this molecule. Calculating WBO..."
            )

        rotors_wbo = {}

        for bond_indices in rotor_bonds:
            bond = molecule.get_bond_between(
                get_atom_index(molecule, bond_indices[0]),
                get_atom_index(molecule, bond_indices[1]),
            )

            rotors_wbo[bond_indices] = bond.fractional_bond_order

        return rotors_wbo

    @classmethod
    def _compare_wbo(
        cls, fragment: Molecule, bond_tuple: BondTuple, parent_wbo: float, **kwargs
    ) -> float:
        """Compare Wiberg Bond order of rotatable bond in a fragment to the parent.

        Parameters
        ----------
        fragment
            The fragment containing the rotatable bond.
        bond_tuple
            The map indices of the rotatable bond.
        parent_wbo
            The WBO of the parent bond with map indices matching ``bond_tuple``.

        Returns
        -------
            The absolute difference between the fragment and parent WBOs.
        """

        # Create new fragment object because sometimes the molecule created from atom
        # bond set is wonky and then the WBOs are not reproducible

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AtomMappingWarning)

            fragment = Molecule.from_smiles(
                fragment.to_smiles(mapped=True), allow_undefined_stereo=True
            )

        fragment_map = fragment.properties.pop("atom_map", None)

        try:
            fragment = assign_elf10_am1_bond_orders(fragment, **kwargs)

        except RuntimeError:
            # Most of the time it fails because it is either missing parameters or a
            # functional group that should not be fragmented was fragmented
            logger.warning(
                f"Cannot calculate WBO for fragment {fragment.to_smiles()}. Continue "
                f"growing fragment"
            )

            # TODO: handle different kinds of failures instead of just continuing to
            #      grow until the failure goes away. Some fail because there are
            #      functional groups that should not be fragmented.
            return 1.0

        if fragment_map is not None:
            fragment.properties["atom_map"] = fragment_map

        bond = fragment.get_bond_between(
            get_atom_index(fragment, bond_tuple[0]),
            get_atom_index(fragment, bond_tuple[1]),
        )

        fragment_wbo = bond.fractional_bond_order

        return abs(parent_wbo - fragment_wbo)

    @classmethod
    def _build_fragment(
        cls,
        parent: Molecule,
        parent_stereo: Stereochemistries,
        parent_groups: FunctionalGroups,
        parent_rings: RingSystems,
        bond_tuple: BondTuple,
        parent_wbo: float,
        threshold: float,
        heuristic: Heuristic = "path_length",
        cap: bool = True,
        **kwargs,
    ) -> Molecule:
        """Build a fragment around a specified bond.

        Parameters
        ----------
        parent
            The original molecule being fragmented.
        parent_stereo
            The stereochemistry of the parent molecule.
        parent_groups
            A dictionary of the functional groups on the molecule which should not
            be fragmented.
        parent_rings
            A dictionary of the ring systems in the molecule which should not
            be fragmented.
        bond_tuple
            The map indices specifying which bond to build the fragment around.
        parent_wbo
            The WBO of the parent bond with map indices matching ``bond_tuple``.
        threshold
            The threshold for the central bond WBO. If the fragment WBO is below this
            threshold, fragmenter will grow out the fragment one bond at a time via the
            path specified by the heuristic option
        heuristic
            The heuristic to use when building the fragment.
        cap
            Whether to cap open valences.

        Returns
        -------
            The built fragment molecule.
        """

        atoms, bonds = cls._get_torsion_quartet(parent, bond_tuple)
        atoms, bonds = cls._get_ring_and_fgroups(
            parent, parent_groups, parent_rings, atoms, bonds
        )

        # Cap open valence
        if cap:
            atoms, bonds = cls._cap_open_valence(parent, parent_groups, atoms, bonds)

        fragment, has_new_stereocenter = cls._atom_bond_set_to_mol(
            parent, parent_stereo, atoms, bonds
        )

        if has_new_stereocenter:
            wbo_difference = threshold + 1.0
        else:
            wbo_difference = cls._compare_wbo(
                fragment, bond_tuple, parent_wbo, **kwargs
            )

        while fragment is not None and wbo_difference > threshold:
            fragment, has_new_stereocenter = cls._add_next_substituent(
                parent,
                parent_stereo,
                parent_groups,
                parent_rings,
                atoms,
                bonds,
                target_bond=bond_tuple,
                heuristic=heuristic,
            )

            if fragment is None:
                break

            if has_new_stereocenter:
                # For now keep growing the fragment as it is not yet clear how to handle such cases robustly.
                wbo_difference = threshold + 1.0

            else:
                wbo_difference = cls._compare_wbo(
                    fragment, bond_tuple, parent_wbo, **kwargs
                )

        # A work around for a known bug where if stereochemistry changes or gets removed,
        # the WBOs can change more than the threshold (this will sometimes happen if a
        # very small threshold is chosen) and even the parent will have a WBO difference
        # greater than the threshold. In this case, return the molecule
        if fragment is None:
            fragment, _ = cls._atom_bond_set_to_mol(parent, parent_stereo, atoms, bonds)

        return fragment

    @classmethod
    def _select_neighbour_by_path_length(
        cls, molecule: Molecule, atoms: set[int], target_bond: BondTuple
    ) -> tuple[int, BondTuple] | None:
        atom_indices = {get_atom_index(molecule, atom) for atom in atoms}

        atoms_to_add = [
            (atom_index, neighbour.molecule_atom_index)
            for atom_index in atom_indices
            for neighbour in molecule.atoms[atom_index].bonded_atoms
            if neighbour.atomic_number != 1
            and neighbour.molecule_atom_index not in atom_indices
        ]
        map_atoms_to_add = [
            (
                get_map_index(molecule, j),
                (get_map_index(molecule, i), get_map_index(molecule, j)),
            )
            for i, j in atoms_to_add
        ]

        # Compute the distance from each neighbouring atom to each of the atoms in the
        # target bond.
        nx_molecule = molecule.to_networkx()

        target_indices = [get_atom_index(molecule, atom) for atom in target_bond]

        path_lengths = list(
            zip(
                *(
                    (
                        networkx.shortest_path_length(
                            nx_molecule, target_index, neighbour_index
                        )
                        for target_index in target_indices
                    )
                    for atom_index, neighbour_index in atoms_to_add
                )
            )
        )

        if len(path_lengths) == 0:
            return None
        path_lengths_1, path_lengths_2 = path_lengths

        reverse = False

        min_path_length_1 = min(path_lengths_1)
        min_path_length_2 = min(path_lengths_2)

        if min_path_length_1 < min_path_length_2:
            sort_by = path_lengths_1
        elif min_path_length_2 < min_path_length_1:
            sort_by = path_lengths_2

        else:
            # If there are multiple neighbouring atoms the same path length away
            # from the target bond fall back to sorting by the WBO.
            map_atoms_to_add = [
                map_tuple
                for map_tuple, *path_length_tuple in zip(
                    map_atoms_to_add, path_lengths_1, path_lengths_2
                )
                if min_path_length_1 in path_length_tuple
            ]

            sort_by = [
                molecule.get_bond_between(
                    get_atom_index(molecule, neighbour_bond[0]),
                    get_atom_index(molecule, neighbour_bond[1]),
                ).fractional_bond_order
                for _, neighbour_bond in map_atoms_to_add
            ]

            reverse = True

        sorted_atoms = [
            a for _, a in sorted(zip(sort_by, map_atoms_to_add), reverse=reverse)
        ]

        return None if len(sorted_atoms) == 0 else sorted_atoms[0]

    @classmethod
    def _select_neighbour_by_wbo(
        cls, molecule: Molecule, atoms: set[int]
    ) -> tuple[int, BondTuple] | None:
        """A function which return those atoms which neighbour those in the ``atoms``
        list sorted by the WBO of the bond between the input atom and the neighbouring
        atom from largest to smallest.

        Parameters
        ----------
        molecule
            The original molecule being fragmented.
        atoms
            The map indices of the atoms currently in the fragment.

        Returns
        -------
            The indices of the atoms to be added to the fragment sorted into the
            order that they should be added in.
        """

        map_bond_orders = {
            (
                get_map_index(molecule, bond.atom1_index),
                get_map_index(molecule, bond.atom2_index),
            ): bond.fractional_bond_order
            for bond in molecule.bonds
            if bond.atom1.atomic_number != 1 and bond.atom2.atomic_number != 1
        }

        neighbour_bond_orders = {
            (bond_order, (map_tuple[1 - i], map_tuple))
            for i in range(2)
            for map_tuple, bond_order in map_bond_orders.items()
            if map_tuple[i] in atoms and map_tuple[1 - i] not in atoms
        }

        sorted_atoms = [
            atom_to_add
            for _, atom_to_add in sorted(
                neighbour_bond_orders, key=lambda x: x[0], reverse=True
            )
        ]

        return None if len(sorted_atoms) == 0 else sorted_atoms[0]

    @classmethod
    def _add_next_substituent(
        cls,
        parent: Molecule,
        parent_stereo: Stereochemistries,
        parent_groups: FunctionalGroups,
        parent_rings: RingSystems,
        atoms: set[int],
        bonds: set[BondTuple],
        target_bond: BondTuple,
        heuristic: Heuristic = "path_length",
    ) -> tuple[Molecule | None, bool]:
        """Expand the fragment to include the next set of substituents / ring systems.

        Parameters
        ----------
        parent
            The original molecule being fragmented.
        parent_stereo
            The stereochemistry of the parent molecule.
        parent_groups
            A dictionary of the functional groups on the molecule which should not
            be fragmented.
        parent_rings
            A dictionary of the ring systems in the molecule which should not
            be fragmented.
        atoms
            The map indices of the atoms currently in the fragment.
        bonds
            The map indices of the bonds currently in the fragment.
        target_bond
            The bond the fragment is being built around.
        heuristic
            How to add substituents. The choices are `'path_length'` or `'wbo'`

        Returns
        -------
            The expanded fragment, or None if the fragment already includes the
            entire parent.
        """

        # Select the next atom neighbour (and the groups / rings that it is part of)
        # that should be added to the fragment.
        if heuristic == "wbo":
            neighbour_atom_and_bond = cls._select_neighbour_by_wbo(parent, atoms)
        elif heuristic == "path_length":
            neighbour_atom_and_bond = cls._select_neighbour_by_path_length(
                parent, atoms, target_bond
            )
        else:
            raise NotImplementedError(
                "Only `'wbo'` and `'path_length'` are supported heuristics."
            )

        if neighbour_atom_and_bond is None:
            return None, False

        neighbour_atom, neighbour_bond = neighbour_atom_and_bond

        # If the neighbour to include is part of a functional group / ring system
        # the entire group should be included in the fragment.
        for group, group_atoms in parent_groups.items():
            if neighbour_atom not in group_atoms[0]:
                continue

            atoms.update(parent_groups[group][0])
            bonds.update(parent_groups[group][1])

        for ring_index, ring_atoms in parent_rings.items():
            if neighbour_atom not in ring_atoms[0]:
                continue

            atoms.update(parent_rings[ring_index][0])
            bonds.update(parent_rings[ring_index][1])

        atoms.add(neighbour_atom)
        bonds.add(neighbour_bond)

        # Check new WBO
        return cls._atom_bond_set_to_mol(parent, parent_stereo, atoms, bonds)


class PfizerFragmenter(Fragmenter):
    """Fragment engine for fragmenting molecules using Pfizer's protocol
    (doi: 10.1021/acs.jcim.9b00373)
    """

    scheme: Literal["Pfizer"] = "Pfizer"

    def _fragment(
        self, molecule: Molecule, target_bond_smarts: list[str] | None
    ) -> FragmentationResult:
        """Fragments a molecule according to Pfizer protocol."""

        (
            parent,
            parent_stereo,
            parent_groups,
            parent_rings,
        ) = self._prepare_molecule(molecule, self.functional_groups, False)

        rotatable_bonds = self.find_rotatable_bonds(parent, target_bond_smarts)

        fragments = {}

        for bond in rotatable_bonds:
            atoms, bonds = self._get_torsion_quartet(parent, bond)

            atoms, bonds = self._get_ring_and_fgroups(
                parent, parent_groups, parent_rings, atoms, bonds
            )

            atoms, bonds = self._cap_open_valence(parent, parent_groups, atoms, bonds)

            fragments[bond], _ = self._atom_bond_set_to_mol(
                parent, parent_stereo, atoms, bonds
            )
        return FragmentationResult(
            parent_smiles=parent.to_smiles(mapped=True),
            fragments=[
                Fragment(smiles=fragment.to_smiles(mapped=True), bond_indices=bond)
                for bond, fragment in fragments.items()
            ],
            provenance=self._default_provenance(),
        )



import warnings
from contextlib import nullcontext

import pytest
from openff.fragmenter.utils import (
    default_functional_groups,
    get_atom_index,
    get_map_index,
)
from openff.toolkit.topology import Molecule
from openff.toolkit.utils.exceptions import AtomMappingWarning


def test_default_functional_groups():
    """Tests that the `default_functional_groups` utility returns correctly."""

    smarts = default_functional_groups()

    assert "hydrazine" in smarts
    assert smarts["hydrazine"] == "[NX3:1][NX3:2]"

    assert "phosphon" not in smarts

    assert len(smarts) == 21


def test_get_map_index():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AtomMappingWarning)

        molecule = Molecule.from_smiles("[C:5]([H:1])([H:2])([H:3])([H:4])")

    assert get_map_index(molecule, 0) == 5


@pytest.mark.parametrize(
    "raise_error, expected_raises",
    [
        (False, nullcontext()),
        (True, pytest.raises(KeyError, match="is not in the atom map")),
    ],
)
def test_get_map_index_error(raise_error, expected_raises):
    molecule = Molecule.from_smiles("C")

    with expected_raises:
        map_index = get_map_index(molecule, 0, error_on_missing=raise_error) == 5
        assert map_index == 0


def test_get_atom_index():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AtomMappingWarning)

        molecule = Molecule.from_smiles("[C:5]([H:1])([H:2])([H:3])([H:4])")

    assert get_atom_index(molecule, 5) == 0



import pytest
from openff.fragmenter.states import _enumerate_stereoisomers
from openff.toolkit.topology import Molecule


@pytest.mark.parametrize(
    "smiles, force_flip, expected",
    [
        (
            "FC(Br)(Cl)[C@@H](Br)(Cl)",
            True,
            [
                "F[C@](Br)(Cl)[C@@H](Br)(Cl)",
                "F[C@](Br)(Cl)[C@H](Br)(Cl)",
                "F[C@@](Br)(Cl)[C@@H](Br)(Cl)",
                "F[C@@](Br)(Cl)[C@H](Br)(Cl)",
            ],
        ),
        (
            "FC(Br)(Cl)[C@@H](Br)(Cl)",
            False,
            ["F[C@](Br)(Cl)[C@@H](Br)(Cl)", "F[C@@](Br)(Cl)[C@@H](Br)(Cl)"],
        ),
        ("FC(Cl)Br", False, ["F[C@H](Cl)Br", "F[C@@H](Cl)Br"]),
        ("F[C@H](Cl)Br", True, ["F[C@H](Cl)Br", "F[C@@H](Cl)Br"]),
        ("F[C@H](Cl)Br", False, ["F[C@H](Cl)Br"]),
        ("ClC=CBr", False, [r"Cl/C=C/Br", r"Cl\C=C/Br"]),
        ("Cl/C=C/Br", False, [r"Cl/C=C/Br"]),
        ("Cl/C=C/Br", True, [r"Cl/C=C/Br", r"Cl\C=C/Br"]),
    ],
)
def test_enumerate_stereoisomers(smiles, force_flip, expected):
    molecule = Molecule.from_smiles(smiles, allow_undefined_stereo=True)

    stereoisomers = _enumerate_stereoisomers(molecule, force_flip=force_flip)

    assert len(stereoisomers) == len(expected)

    expected = {
        Molecule.from_smiles(stereoisomer, allow_undefined_stereo=True).to_smiles(
            explicit_hydrogens=False, isomeric=True, mapped=False
        )
        for stereoisomer in expected
    }
    actual = {
        stereoisomer.to_smiles(explicit_hydrogens=False, isomeric=True, mapped=False)
        for stereoisomer in stereoisomers
    }

    assert expected == actual



"""
Unit and regression test for the fragmenter.fragmenter module.
"""

import logging

import numpy
import pytest
from openff.fragmenter._tests import does_not_raise
from openff.fragmenter._tests.utils import (
    key_smarts_to_map_indices,
    smarts_set_to_map_indices,
    value_smarts_to_map_indices,
)
from openff.fragmenter.chemi import assign_elf10_am1_bond_orders, smiles_to_molecule
from openff.fragmenter.fragment import (
    FragmentationResult,
    Fragmenter,
    PfizerFragmenter,
    WBOFragmenter,
)
from openff.fragmenter.utils import (
    default_functional_groups,
    get_atom_index,
    get_map_index,
)
from openff.toolkit.topology import Molecule
from openff.toolkit.utils import (
    GLOBAL_TOOLKIT_REGISTRY,
    RDKitToolkitWrapper,
    ToolkitRegistry,
)


@pytest.mark.parametrize(
    "smiles, output",
    [
        (
            "OC1(CN(C1)C(=O)C1=C(NC2=C(F)C=C(I)C=C2)C(F)=C(F)C=C1)[C@@H]1CCCCN1",
            {"[#7r6]-[#6:1]-[#6!r6]": "S"},
        ),
        (
            "OC1(CN(C1)C(=O)C1=C(NC2=C(F)C=C(I)C=C2)C(F)=C(F)C=C1)[C@H]1CCCCN1",
            {"[#7r6]-[#6:1]-[#6!r6]": "R"},
        ),
        (r"C\C=C\C", {"[#6:1]=[#6:2]": "E"}),
        (r"C/C=C\C", {"[#6:1]=[#6:2]": "Z"}),
    ],
)
def test_find_stereo(smiles, output):
    molecule = smiles_to_molecule(smiles, add_atom_map=True)

    actual_stereo = Fragmenter._find_stereo(molecule)
    expected_stereo = key_smarts_to_map_indices(output, molecule)

    assert actual_stereo == expected_stereo


@pytest.mark.parametrize(
    "smiles, fragment_smiles, output, warning",
    [
        ("C[C@@](F)(Cl)I", "C[C@@](F)(Cl)I", True, None),
        ("C[C@@](F)(Cl)I", "C[C@](F)(Cl)I", False, "Stereochemistry for atom "),
        ("C[C@@](F)(Cl)F", "C[C@@](F)(Cl)I", False, "A new stereocenter formed at"),
        (r"C\C=C\C", r"C\C=C\C", True, None),
        (r"C/C=C\C", r"C\C=C\C", False, "Stereochemistry for bond"),
        ("CC=C(C)C", r"C\C=C(C)\CC", False, "A new chiral bond formed at"),
    ],
)
def test_check_stereo(smiles, fragment_smiles, output, warning, caplog):
    parent = smiles_to_molecule(smiles, add_atom_map=True)
    fragment = smiles_to_molecule(fragment_smiles, add_atom_map=True)

    parent_stereo = Fragmenter._find_stereo(parent)

    with caplog.at_level(logging.WARNING):
        assert Fragmenter._check_stereo(fragment, parent_stereo) == output

    if warning is None:
        assert len(caplog.records) == 0
    else:
        assert len(caplog.records) == 1
        assert caplog.records[0].message.startswith(warning)


@pytest.mark.parametrize(
    "smiles, fragment_smiles",
    [
        ("C[C@@](F)(Cl)I", "C[C@@](F)(Cl)I"),
        ("C[C@@](F)(Cl)I", "C[C@](F)(Cl)I"),
        (r"C\C=C\C", r"C\C=C\C"),
        (r"C/C=C\C", r"C\C=C\C"),
    ],
)
def test_fix_stereo(smiles, fragment_smiles):
    parent_stereo = Fragmenter._find_stereo(smiles_to_molecule(smiles, True))

    fragment = smiles_to_molecule(fragment_smiles, add_atom_map=True)
    fixed = Fragmenter._fix_stereo(fragment, parent_stereo)

    assert Fragmenter._check_stereo(fixed, parent_stereo) is True


def test_find_functional_groups(potanib):
    found_groups = Fragmenter._find_functional_groups(
        molecule=potanib,
        functional_groups={
            # Taken from the old `fgroup_smarts_comb.yml` file.
            "amide": "[NX3R0:1][CX3:2](=[OX1:3])",
            "tri_halide": "[#6:1]([#9:2])([#9:3])([#9:4])",
            "carbonyl": "[CX3R0:1]=[OX1:2]",
            "alkyne": "[CX2:1]#[CX2:2]",
        },
    )

    expected_functional_groups = value_smarts_to_map_indices(
        {
            "alkyne_0": {"[#6r6]-[#6:1]#[#6]", "[#6r5]-[#6:1]#[#6]"},
            "carbonyl_0": {"[#6:1]=[#8]", "[#6]=[#8:1]"},
            "amide_0": {"[#6:1]=[#8]", "[#6]=[#8:1]", "[#7:1]-[#6]=[#8]"},
            "tri_halide_0": {"[#6:1]-[#9]", "[#6]-[#9:1]"},
        },
        potanib,
    )
    actual_functional_groups = {
        name: groups[0] for name, groups in found_groups.items()
    }

    assert actual_functional_groups == expected_functional_groups

    for group in found_groups:
        assert all(
            index in expected_functional_groups[group]
            for bond in found_groups[group][1]
            for index in bond
        )


def test_find_rotatable_bonds_default(abemaciclib):
    rotatable_bonds = Fragmenter.find_rotatable_bonds(abemaciclib, None)
    assert len(rotatable_bonds) == 7

    expected_rotatable_bonds = smarts_set_to_map_indices(
        {
            "[#6ar6:1]-[#6ar6:2]",
            "[#6ar6:1]-[#6H2:2]",
            "[#7r6:1]-[#6H2:2]-[#6a]",
            "[#7r6:1]-[#6H2:2]-[#6H3]",
            "[#6ar6:1]-[#7H1:2]-[#6a](:[#7a])(:[#7a])",
            "[#7r5:1]-[#6:2](-[#6H3])(-[#6H3])",
            "[#6ar6:1]-[#7H1:2]-[#6a]:[#6a]",
        },
        abemaciclib,
    )

    assert {*rotatable_bonds} == expected_rotatable_bonds


@pytest.mark.parametrize(
    "smarts, expected_raises",
    [
        (["[#6:1]-[#6:2]"], does_not_raise()),
        (
            ["[#6]-[#6]"],
            pytest.raises(ValueError, match="The `target_bond_smarts` pattern "),
        ),
    ],
)
def test_find_rotatable_bonds_custom(smarts, expected_raises):
    ethane = Molecule.from_smiles("CC")
    ethane.properties["atom_map"] = {i: i + 1 for i in range(ethane.n_atoms)}

    with expected_raises:
        rotatable_bonds = Fragmenter.find_rotatable_bonds(ethane, smarts)
        assert len(rotatable_bonds) == 1

        assert rotatable_bonds == [(1, 2)]


def test_atom_bond_set_to_mol(abemaciclib):
    molecule = smiles_to_molecule(abemaciclib.to_smiles(mapped=False), True)

    atoms = {
        get_map_index(molecule, atom_index)
        for match in molecule.chemical_environment_matches(
            "[C:1][C:2][N:3]1[C:4][C:5][N:6][C:7][C:8]1"
        )
        for atom_index in match
    }

    bonds = {
        (
            get_map_index(molecule, bond.atom1_index),
            get_map_index(molecule, bond.atom2_index),
        )
        for bond in molecule.bonds
        if get_map_index(molecule, bond.atom1_index) in atoms
        and get_map_index(molecule, bond.atom2_index) in atoms
    }

    fragment, _ = Fragmenter._atom_bond_set_to_mol(
        molecule, {}, atoms=atoms, bonds=bonds
    )

    for bond in fragment.bonds:
        if bond.atom1.atomic_number == 1 or bond.atom2.atomic_number == 1:
            continue

        map_index_1 = get_map_index(fragment, bond.atom1_index)
        map_index_2 = get_map_index(fragment, bond.atom2_index)

        assert tuple(sorted((map_index_1, map_index_2))) in bonds


@pytest.mark.parametrize(
    "input_smiles, expected_n_atoms, expected_n_bonds", [("CCCCCCC", 14, 20)]
)
def test_get_torsion_quartet(input_smiles, expected_n_atoms, expected_n_bonds):
    molecule = smiles_to_molecule(input_smiles, True)

    bond_match = molecule.chemical_environment_matches("C[C:1][C:2]CCCC")[0]

    atoms, bonds = Fragmenter._get_torsion_quartet(
        molecule,
        (
            get_map_index(molecule, bond_match[0]),
            get_map_index(molecule, bond_match[1]),
        ),
    )

    # This also includes explicit hydrogen
    assert len(atoms) == expected_n_atoms
    assert len(bonds) == expected_n_bonds


@pytest.mark.parametrize(
    "input_smiles, n_ring_systems", [("CCCC", 0), ("c1ccccc1", 1), ("c1ccccc1C", 1)]
)
def test_find_ring_systems(input_smiles, n_ring_systems):
    ring_systems = Fragmenter._find_ring_systems(
        smiles_to_molecule(input_smiles, True), {}
    )

    assert len(ring_systems) == n_ring_systems


@pytest.mark.parametrize(
    "keep_non_rotor_ring_substituents, n_output", [(True, 7), (False, 6)]
)
def test_keep_non_rotor(keep_non_rotor_ring_substituents, n_output):
    ring_systems = Fragmenter._find_ring_systems(
        smiles_to_molecule("c1ccccc1C", True),
        {},
        keep_non_rotor_ring_substituents=keep_non_rotor_ring_substituents,
    )

    assert len(ring_systems[1][0]) == n_output


@pytest.mark.parametrize(
    "input_smiles, bond_smarts, expected",
    [
        ("CCCCCCC", "C[C:1][C:2]CCCC", True),
        ("CCCCc1ccccc1", "[#6H3]-[#6H2:1]-[#6H2:2]", False),
    ],
)
def test_get_ring_and_fgroup(input_smiles, bond_smarts, expected):
    molecule, _, functional_groups, ring_systems = Fragmenter._prepare_molecule(
        smiles_to_molecule(input_smiles, True), default_functional_groups(), False
    )

    # noinspection PyTypeChecker
    atoms, bonds = Fragmenter._get_torsion_quartet(
        molecule,
        tuple(
            get_map_index(molecule, i)
            for i in molecule.chemical_environment_matches(bond_smarts)[0]
        ),
    )

    bonds = {tuple(sorted(bond)) for bond in bonds}

    l_atoms = len(atoms)
    l_bonds = len(bonds)

    atoms_2, bonds_2 = Fragmenter._get_ring_and_fgroups(
        molecule, functional_groups, ring_systems, atoms, bonds
    )

    assert (l_atoms == len(atoms_2)) == expected
    assert (l_bonds == len(bonds_2)) == expected


@pytest.mark.parametrize(
    "input_smiles, bond_smarts, expected_pattern",
    [
        (
            "CCc1ccccc1C=O",
            "C[CH2:2][c:1]1ccccc1C=O",
            "[CH3:8][CH2:9][c:6]1[cH:4][cH:2][cH:1][cH:3][c:5]1[CH:7]=[O:10]",
        ),
        (
            "c1ccc(cc1)c2ccccc2c3ccccc3",
            "c1cc[c:1](cc1)[c:2]2ccccc2c3ccccc3",
            "[cH:1]1[cH:3][cH:9][c:15]([cH:10][cH:4]1)[c:17]2[cH:13][cH:7][cH:8]"
            "[cH:14][c:18]2[c:16]3[cH:11][cH:5][cH:2][cH:6][cH:12]3",
        ),
    ],
)
def test_get_ring_and_fgroup_ortho(input_smiles, bond_smarts, expected_pattern):
    """Ensure that FGs and rings attached to ortho groups are correctly
    detected.

    The expected values were generated using fragmenter=0.0.7
    """

    molecule, _, functional_groups, ring_systems = Fragmenter._prepare_molecule(
        smiles_to_molecule(input_smiles, True), default_functional_groups(), False
    )

    bond = tuple(
        get_map_index(molecule, i)
        for i in molecule.chemical_environment_matches(bond_smarts)[0]
    )

    # noinspection PyTypeChecker
    atoms, bonds = Fragmenter._get_torsion_quartet(molecule, bond)
    atoms, bonds = Fragmenter._get_ring_and_fgroups(
        molecule, functional_groups, ring_systems, atoms, bonds
    )

    actual_atoms = {
        map_index
        for map_index in atoms
        if molecule.atoms[get_atom_index(molecule, map_index)].atomic_number != 1
    }
    expected_atoms = {
        get_map_index(molecule, atom_index)
        for match in molecule.chemical_environment_matches(expected_pattern)
        for atom_index in match
    }

    assert actual_atoms == expected_atoms


def test_find_ortho_substituents(dasatanib):
    # TODO: add test to test adding substituent directly bonded to rotatable bond
    ortho_atoms, ortho_bonds = Fragmenter._find_ortho_substituents(
        dasatanib,
        bonds=smarts_set_to_map_indices({"[#6ar6:1]-[#7:2]-[#6]=[#8]"}, dasatanib),
    )

    assert ortho_atoms == smarts_set_to_map_indices(
        {"[#6ar6]-[#7:1]-[#6]=[#8]", "[#6ar6]-[#17:1]", "[#6H3:1]-[#6r6]:[#6r6]"},
        dasatanib,
    )
    assert ortho_bonds == smarts_set_to_map_indices(
        {"[#6ar6:1]-[#7:2]-[#6]=[#8]", "[#6ar6:1]-[#17:2]", "[#6H3:2]-[#6r6:1]:[#6r6]"},
        dasatanib,
    )


def test_cap_open_valance():
    molecule, _, functional_groups, ring_systems = Fragmenter._prepare_molecule(
        smiles_to_molecule("CNCCc1ccccc1", True), default_functional_groups(), False
    )

    expected_atom = get_map_index(
        molecule,
        molecule.chemical_environment_matches("[#7]-[#6H3:1]")[0][0],
    )

    # noinspection PyTypeChecker
    atoms, bonds = Fragmenter._get_torsion_quartet(
        molecule,
        tuple(
            get_map_index(molecule, i)
            for i in molecule.chemical_environment_matches("[#6a:1]-[#6H2:2]")[0]
        ),
    )
    atoms, bonds = Fragmenter._get_ring_and_fgroups(
        molecule, functional_groups, ring_systems, atoms, bonds
    )

    # Remove the cap atom from the current list to make sure it gets included during
    # capping.
    atoms -= {expected_atom}

    atoms, _ = Fragmenter._cap_open_valence(molecule, functional_groups, atoms, bonds)

    # Check that carbon bonded to N was added
    assert expected_atom in atoms


def test_prepare_molecule():
    molecule, stereo, functional_groups, ring_systems = Fragmenter._prepare_molecule(
        Molecule.from_smiles("C[C@H](O)C1CCCC1"), {"alcohol": "[#6:1]-[#8X2H1:2]"}, True
    )

    assert "atom_map" in molecule.properties

    assert {*stereo.values()} == {"S"}
    assert len(stereo) == 1

    assert "alcohol_0" in functional_groups
    assert len(functional_groups["alcohol_0"][0]) == 2
    assert len(functional_groups["alcohol_0"][1]) == 1

    assert len(ring_systems) == 1
    assert len(ring_systems[1][0]) == 5


@pytest.mark.parametrize(
    "toolkit_registry, expected_provenance",
    [
        (
            None,
            [
                toolkit.__class__.__name__
                for toolkit in GLOBAL_TOOLKIT_REGISTRY.registered_toolkits
            ],
        ),
        (RDKitToolkitWrapper(), ["RDKitToolkitWrapper"]),
        (ToolkitRegistry([RDKitToolkitWrapper]), ["RDKitToolkitWrapper"]),
    ],
)
def test_fragmenter_provenance(toolkit_registry, expected_provenance):
    class DummyFragmenter(Fragmenter):
        def _fragment(
            self, molecule: Molecule, target_bond_smarts: str
        ) -> FragmentationResult:
            return FragmentationResult(
                parent_smiles="[He:1]", fragments=[], provenance={}
            )

    result = DummyFragmenter().fragment(
        Molecule.from_smiles("[He]"), ["[*:1]~[*:2]"], toolkit_registry
    )

    assert "toolkits" in result.provenance
    assert [name for name, _ in result.provenance["toolkits"]] == expected_provenance

    assert "options" in result.provenance
    assert result.provenance["options"]["target_bond_smarts"] == ["[*:1]~[*:2]"]


def test_wbo_fragment():
    """Test build fragment"""

    result = WBOFragmenter().fragment(Molecule.from_smiles("CCCCC"))

    assert len(result.fragments) == 2
    assert {fragment.bond_indices for fragment in result.fragments} == {(3, 5), (4, 5)}


def test_keep_track_of_map():
    result = WBOFragmenter().fragment(Molecule.from_smiles("CCCCC"))

    assert all(
        "atom_map" in fragment.molecule.properties for fragment in result.fragments
    )


def test_get_rotor_wbo():
    molecule = smiles_to_molecule("CCCC", True)

    for bond in molecule.bonds:
        bond.fractional_bond_order = 0.986

    expected_bonds = {
        (
            get_map_index(molecule, match[0]),
            get_map_index(molecule, match[1]),
        )
        for match in molecule.chemical_environment_matches("[#6:1]-[#6:2]")
    }

    rotors_wbo = WBOFragmenter._get_rotor_wbo(
        molecule, WBOFragmenter.find_rotatable_bonds(molecule, None)
    )

    assert len(rotors_wbo) == 1

    rotor_index = next(iter(rotors_wbo))

    assert rotor_index in expected_bonds
    assert numpy.isclose(rotors_wbo[rotor_index], 0.986, atol=0.001)


def test_compare_wbo():
    parent = assign_elf10_am1_bond_orders(smiles_to_molecule("CCCC", add_atom_map=True))
    rotors_wbo = WBOFragmenter._get_rotor_wbo(
        parent, WBOFragmenter.find_rotatable_bonds(parent, None)
    )

    fragment = smiles_to_molecule("CCCC", add_atom_map=True)

    for bond_tuple, value in rotors_wbo.items():
        assert numpy.isclose(
            WBOFragmenter._compare_wbo(fragment, bond_tuple, value),
            0.0,
            atol=1.0e-6,
        )
        assert numpy.isclose(
            WBOFragmenter._compare_wbo(fragment, bond_tuple, value + 1.0),
            1.0,
            atol=1.0e-6,
        )


def test_build_fragment():
    parent = assign_elf10_am1_bond_orders(smiles_to_molecule("CCCCCC", True))
    rotors_wbo = WBOFragmenter._get_rotor_wbo(
        parent, WBOFragmenter.find_rotatable_bonds(parent, None)
    )

    fragments = {
        bond: WBOFragmenter._build_fragment(
            parent, {}, {}, {}, bond, parent_wbo, threshold=0.05
        )
        for bond, parent_wbo in rotors_wbo.items()
    }

    assert len(fragments) == 3

    assert (
        fragments[(3, 5)].to_smiles(explicit_hydrogens=False, mapped=False) == "CCCCC"
    )
    assert (
        fragments[(4, 6)].to_smiles(explicit_hydrogens=False, mapped=False) == "CCCCC"
    )
    assert (
        fragments[(5, 6)].to_smiles(explicit_hydrogens=False, mapped=False) == "CCCCCC"
    )


@pytest.mark.parametrize(
    "input_smiles, n_output",
    [
        (
            "[H:42][N:19]([C@H:11]1[C@@H:13]2[N:18]([C:7]1=[O:22])[CH2:12][C:14]"
            "([S:26]2)([CH3:15])[CH3:16])[C:9](=[O:24])[CH3:17]",
            8,
        )
    ],
)
def test_ring_fgroups(input_smiles, n_output):
    parent = smiles_to_molecule(input_smiles, True)
    parent_groups = Fragmenter._find_functional_groups(
        parent, default_functional_groups()
    )

    parent_rings = Fragmenter._find_ring_systems(parent, parent_groups)

    assert len(parent_rings[1][0]) == n_output


def test_add_substituent():
    result = WBOFragmenter().fragment(Molecule.from_smiles("CCCCCC"))

    fragment = result.fragments_by_bond[(3, 5)].molecule

    assert fragment.to_smiles(mapped=False, explicit_hydrogens=False) == "CCCCC"

    atoms = {
        get_map_index(fragment, i)
        for i in range(fragment.n_atoms)
        if fragment.atoms[i].atomic_number != 1
    }

    bonds = {
        (
            get_map_index(fragment, bond.atom1_index),
            get_map_index(fragment, bond.atom2_index),
        )
        for bond in fragment.bonds
        if bond.atom1.atomic_number != 1 and bond.atom2.atomic_number != 1
    }

    fragment, _ = WBOFragmenter._add_next_substituent(
        result.parent_molecule, {}, {}, {}, atoms, bonds, target_bond=(3, 5)
    )

    assert fragment.to_smiles(mapped=False, explicit_hydrogens=False) == "CCCCCC"


def test_issue_177():
    """
    There was an where the path search gets tripped up when there are two paths
    of equal length leading to a 3-membered ring. This test ensures Fragmenter
    runs on a minimal reproducing example of this problem.
    More info at https://github.com/openforcefield/openff-fragmenter/issues/177
    """
    WBOFragmenter().fragment(Molecule.from_smiles("C1CC1[C@H]2C[C@H]3C[C@H]3C2"))


@pytest.mark.parametrize("input_smiles, n_output", [("CCCCCCC", 4)])
def test_pfizer_fragmenter(input_smiles, n_output):
    result = PfizerFragmenter().fragment(Molecule.from_smiles(input_smiles))
    assert len(result.fragments) == n_output



"""functions to manipulate, read and write OpenEye and Psi4 molecules"""

import copy
import logging

import networkx
import numpy
from openff.fragmenter.utils import get_atom_index, get_map_index
from openff.toolkit.topology import Molecule
from openff.toolkit.utils import LicenseError, ToolkitUnavailableException
from openff.utilities import MissingOptionalDependencyError, requires_oe_module

logger = logging.getLogger(__name__)


def assign_elf10_am1_bond_orders(
    molecule: Molecule, max_confs: int = 800, rms_threshold: float = 1.0
) -> Molecule:
    """Generate ELF10 AM1 WBOs for a molecule.

    Parameters
    ----------
    molecule
        The molecule to compute WBOs for.
    max_confs : int, optional, default=800
        Max number of conformers to generate
    rms_threshold
        The minimum RMS value [Angstrom] at which two conformers are considered redundant
        and one is deleted.

    Returns
    -------
        A new molecule which contains ELF10 AM1 Wiberg Bond orders.
    """

    molecule = copy.deepcopy(molecule)

    # Store the atom map separately in case it gets removed by a TK.
    atom_map = molecule.properties.pop("atom_map", None)

    # Generate a set of ELF10 conformers.
    molecule = _generate_conformers(molecule, max_confs, rms_threshold)
    molecule.apply_elf_conformer_selection()

    per_conformer_bond_orders = []

    for conformer in molecule.conformers:
        molecule.assign_fractional_bond_orders("am1-wiberg", use_conformers=[conformer])

        per_conformer_bond_orders.append(
            [bond.fractional_bond_order for bond in molecule.bonds]
        )

    bond_orders = [*numpy.mean(per_conformer_bond_orders, axis=0)]

    for bond, bond_order in zip(molecule.bonds, bond_orders):
        bond.fractional_bond_order = bond_order

    # Restore the atom map.
    if atom_map is not None:
        molecule.properties["atom_map"] = atom_map

    return molecule


def _generate_conformers(
    molecule: Molecule, max_confs: int = 800, rms_threshold: float = 1.0
) -> Molecule:
    """Generate conformations for the supplied molecule.

    Parameters
    ----------
    molecule
        Molecule for which to generate conformers
    max_confs
        Max number of conformers to generate.
    rms_threshold
        The minimum RMS value [Angstrom] at which two conformers are considered redundant
        and one is deleted.

    Returns
    -------
        A new multi-conformer molecule with up to max_confs conformers.
    """

    from openff.units import unit

    molecule = copy.deepcopy(molecule)

    # Store the atom map separately in case it gets removed / mangled by a TK.
    atom_map = molecule.properties.pop("atom_map", None)

    # Canonically order the atoms in the molecule before generating the conformer.
    # This helps ensure the same conformers are generated for the same molecules
    # independently of their atom order.
    canonical_molecule = molecule.canonical_order_atoms()

    canonical_molecule.generate_conformers(
        n_conformers=max_confs, rms_cutoff=rms_threshold * unit.angstrom
    )

    _, canonical_map = Molecule.are_isomorphic(
        canonical_molecule, molecule, return_atom_map=True
    )

    molecule = canonical_molecule.remap(canonical_map)

    # Restore the atom map.
    if atom_map is not None:
        molecule.properties["atom_map"] = atom_map

    return molecule


def find_ring_systems(molecule: Molecule) -> dict[int, int]:
    """This function attempts to find all ring systems (see [1] for more details) in
    a given molecule.

    The method first attempts to determine which atoms and bonds are part of rings
    by matching the `[*:1]@[*:2]` SMIRKS pattern against the molecule.

    The matched bonds are then used to construct a graph (using ``networkx``), from which
    the ring systems are identified as those sets of atoms which are 'connected'
    together (using ``connected_components``) by at least one path.

    Parameters
    ----------
    molecule:
        The molecule to search for ring systems.

    Notes
    -----
    * Two molecular rings with only one common atom (i.e. spiro compounds) are
      considered to be part of the same ring system.

    References
    ----------
    [1] `Ring Perception <https://docs.eyesopen.com/toolkits/python/oechemtk/ring.html>`_

    Returns
    -------
        The index of which ring system each atom belongs to. Only ring atoms are
        included in the returned dictionary.
    """

    # Find the ring atoms
    ring_atom_index_pairs = {
        tuple(sorted(pair))
        for pair in molecule.chemical_environment_matches("[*:1]@[*:2]")
    }

    # Construct a networkx graph from the found ring bonds.
    graph = networkx.Graph()

    for atom_index_pair in ring_atom_index_pairs:
        graph.add_edge(*atom_index_pair)

    ring_systems = {}

    for i, ring_system in enumerate(networkx.connected_components(graph)):
        for atom_index in ring_system:
            assert atom_index not in ring_systems
            ring_systems[atom_index] = i + 1

    return ring_systems


@requires_oe_module("oechem")
def _find_oe_stereocenters(
    molecule: Molecule,
) -> tuple[list[int], list[tuple[int, int]]]:
    """A method which returns the of the stereogenic atom and bonds of a molecule
    using the OpenEye toolkit.

    Parameters
    ----------
    molecule
        The molecule whose stereocenters should be returned.

    Notes
    -----
    * This method currently deals with OE directly and should be removed once
      the API points suggested in openff-toolkit issue #903 have been added.

    Returns
    -------
        The indices of the stereogenic atoms in the molecule and the indices of the
        atoms involved in stereogenic bonds in the molecule.
    """

    from openeye import oechem

    oe_molecule = molecule.to_openeye()
    oechem.OEPerceiveChiral(oe_molecule)

    stereogenic_atoms = {
        atom.GetIdx() for atom in oe_molecule.GetAtoms() if atom.IsChiral()
    }

    stereogenic_bonds = {
        (bond.GetBgnIdx(), bond.GetEndIdx())
        for bond in oe_molecule.GetBonds()
        if bond.IsChiral()
    }

    return sorted(stereogenic_atoms), sorted(stereogenic_bonds)


def _find_rd_stereocenters(
    molecule: Molecule,
) -> tuple[list[int], list[tuple[int, int]]]:
    """A method which returns the of the stereogenic atom and bonds of a molecule
    using the RDKit.

    Parameters
    ----------
    molecule
        The molecule whose stereocenters should be returned.

    Notes
    -----
    * This method currently deals with RDKit directly and should be removed once
      the API points suggested in openff-toolkit issue #903 have been added.

    Returns
    -------
        The indices of the stereogenic atoms in the molecule and the indices of the
        atoms involved in stereogenic bonds in the molecule.
    """

    from rdkit import Chem

    rd_molecule = molecule.to_rdkit()

    Chem.AssignStereochemistry(
        rd_molecule, cleanIt=True, force=True, flagPossibleStereoCenters=True
    )

    stereogenic_atoms = {
        atom.GetIdx()
        for atom in rd_molecule.GetAtoms()
        if atom.HasProp("_ChiralityPossible")
    }

    # Clear any previous assignments on the bonds, since FindPotentialStereo
    # may not overwrite it
    for bond in rd_molecule.GetBonds():
        bond.SetStereo(Chem.BondStereo.STEREONONE)

    Chem.FindPotentialStereoBonds(rd_molecule, cleanIt=True)

    stereogenic_bonds = {
        (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())
        for bond in rd_molecule.GetBonds()
        if bond.GetStereo() == Chem.BondStereo.STEREOANY
    }

    return sorted(stereogenic_atoms), sorted(stereogenic_bonds)


def find_stereocenters(molecule: Molecule) -> tuple[list[int], list[tuple[int, int]]]:
    """A method which returns the of the stereogenic atom and bonds of a molecule.

    Parameters
    ----------
    molecule
        The molecule whose stereocenters should be returned.

    Notes
    -----
    * This method currently deals with OE and RDKit molecules and should be updated once
      the API points suggested in openff-toolkit issue #903 have been added.

    Returns
    -------
        The indices of the stereogenic atoms in the molecule and the indices of the
        atoms involved in stereogenic bonds in the molecule.
    """

    try:
        stereogenic_atoms, stereogenic_bonds = _find_oe_stereocenters(molecule)
    except (
        ModuleNotFoundError,
        ToolkitUnavailableException,
        LicenseError,
        MissingOptionalDependencyError,
    ):
        stereogenic_atoms, stereogenic_bonds = _find_rd_stereocenters(molecule)

    return stereogenic_atoms, stereogenic_bonds


def _extract_rd_fragment(
    molecule: Molecule, atom_indices: set[int], bond_indices: set[tuple[int, int]]
) -> Molecule:
    from rdkit import Chem

    rd_molecule = Chem.RWMol(molecule.to_rdkit())
    rd_atoms_by_map: dict[int, Chem.Atom] = {}

    # Restore the map indices as to_rdkit does not automatically add them.
    for atom in rd_molecule.GetAtoms():
        atom.SetAtomMapNum(get_map_index(molecule, atom.GetIdx()))

        rd_atoms_by_map[atom.GetAtomMapNum()] = atom

    atoms_to_use = [get_atom_index(molecule, i) for i in atom_indices]
    bonds_to_use = [
        rd_molecule.GetBondBetweenAtoms(
            get_atom_index(molecule, pair[0]), get_atom_index(molecule, pair[1])
        ).GetIdx()
        for pair in bond_indices
    ]

    # Make sure to include any Hs bonded to the included atom set otherwise radicals
    # will form.
    for map_index in atom_indices:
        for neighbour in rd_atoms_by_map[map_index].GetNeighbors():
            if (
                neighbour.GetAtomicNum() != 1
                or neighbour.GetAtomMapNum() < 1
                or neighbour.GetAtomMapNum() in atom_indices
            ):
                continue

            atoms_to_use.append(neighbour.GetIdx())
            bonds_to_use.append(
                rd_molecule.GetBondBetweenAtoms(
                    rd_atoms_by_map[map_index].GetIdx(), neighbour.GetIdx()
                ).GetIdx()
            )

    # Add additional hydrogens to atoms where the total valence will change likewise to
    # ensure the valence does not change.
    rd_atoms_by_index = {atom.GetIdx(): atom for atom in rd_molecule.GetAtoms()}

    for atom_index in [*atoms_to_use]:
        atom = rd_atoms_by_index[atom_index]

        old_valence = atom.GetTotalValence()
        new_valence = atom.GetTotalValence()

        for neighbour_bond in rd_atoms_by_index[atom_index].GetBonds():
            if (
                neighbour_bond.GetBeginAtomIdx() in atoms_to_use
                and neighbour_bond.GetEndAtomIdx() in atoms_to_use
            ):
                continue

            new_valence -= neighbour_bond.GetValenceContrib(atom)

        if numpy.isclose(old_valence, new_valence):
            # Skip the cases where the valence won't change
            continue

        if (
            atom.GetAtomicNum() == 6
            and atom.GetIsAromatic()
            and sum(
                1 for bond_tuple in bond_indices if atom.GetAtomMapNum() in bond_tuple
            )
            == 1
        ):
            # This is likely a cap carbon which was retained from an existing ring. It's
            # aromaticity needs to be cleared before calling ``MolFragmentToSmiles``
            # otherwise will (understandably) be confused and throw an exception.
            atom.SetIsAromatic(False)

        # Add a hydrogen to the atom whose valence will change.
        for _ in range(int(numpy.rint(old_valence - new_valence))):
            new_atom = Chem.Atom(1)
            new_atom_index = rd_molecule.AddAtom(new_atom)

            rd_molecule.AddBond(atom_index, new_atom_index)

            new_bond = rd_molecule.GetBondBetweenAtoms(atom_index, new_atom_index)
            new_bond.SetBondType(Chem.BondType.SINGLE)
            new_bond.SetIsAromatic(False)

            atoms_to_use.append(new_atom_index)
            bonds_to_use.append(new_bond.GetIdx())

    fragment_smiles = Chem.MolFragmentToSmiles(rd_molecule, atoms_to_use, bonds_to_use)
    fragment = Molecule.from_smiles(fragment_smiles, allow_undefined_stereo=True)

    return fragment


@requires_oe_module("oechem")
def _extract_oe_fragment(
    molecule: Molecule, atom_indices: set[int], bond_indices: set[tuple[int, int]]
) -> Molecule:
    from openeye import oechem

    oe_molecule = molecule.to_openeye()

    # Restore the map indices as to_openeye does not automatically add them.
    for atom_index, map_index in molecule.properties["atom_map"].items():
        oe_atom = oe_molecule.GetAtom(oechem.OEHasAtomIdx(atom_index))
        oe_atom.SetMapIdx(map_index)

    # Include any Hs bonded to the included atom set so we can retain their map
    # indices.
    for map_index in {*atom_indices}:
        oe_atom = oe_molecule.GetAtom(oechem.OEHasMapIdx(map_index))

        for neighbour in oe_atom.GetAtoms():
            if (
                neighbour.GetAtomicNum() != 1
                or neighbour.GetMapIdx() < 1
                or neighbour.GetMapIdx() in atom_indices
            ):
                continue

            atom_indices.add(neighbour.GetMapIdx())
            bond_indices.add((map_index, neighbour.GetMapIdx()))

    atom_bond_set = oechem.OEAtomBondSet()

    for map_index in atom_indices:
        atom = oe_molecule.GetAtom(oechem.OEHasMapIdx(map_index))
        atom_bond_set.AddAtom(atom)

    for map_index_1, map_index_2 in bond_indices:
        atom_1 = oe_molecule.GetAtom(oechem.OEHasMapIdx(map_index_1))
        atom_2 = oe_molecule.GetAtom(oechem.OEHasMapIdx(map_index_2))

        bond = oe_molecule.GetBond(atom_1, atom_2)

        if not bond:
            raise ValueError(f"{(map_index_1, map_index_2)} is a disconnected bond")

        atom_bond_set.AddBond(bond)

    atom_predicate = oechem.OEIsAtomMember(atom_bond_set.GetAtoms())
    bond_predicate = oechem.OEIsBondMember(atom_bond_set.GetBonds())

    fragment = oechem.OEMol()
    oechem.OESubsetMol(fragment, oe_molecule, atom_predicate, bond_predicate, True)

    oechem.OEAddExplicitHydrogens(fragment)
    oechem.OEPerceiveChiral(fragment)

    # Always restore map?
    # if restore_maps:
    # In some cases (symmetric molecules) this changes the atom map so skip it
    # restore_atom_map(fragment)
    # atom map should be restored for combinatorial fragmentation
    # Perceive stereo and check that defined stereo did not change
    oechem.OEPerceiveChiral(fragment)
    oechem.OE3DToAtomStereo(fragment)
    oechem.OE3DToBondStereo(fragment)

    return Molecule.from_openeye(fragment, allow_undefined_stereo=True)


def extract_fragment(
    molecule: Molecule, atom_indices: set[int], bond_indices: set[tuple[int, int]]
) -> Molecule:
    """Returns a fragment which contains the specified atoms and bonds of a parent
    molecule

    Parameters
    ----------
    molecule
        The parent molecule being fragmented.
    atom_indices
        The *map* indices of the subset of atoms to retain.
    bond_indices
        The *map* indices of the subset of bonds to retain.
    """

    # Make sure that the bond and atom indices are self consistent and that there
    # are no disconnected bonds.
    if not all(i in atom_indices for map_tuple in bond_indices for i in map_tuple):
        raise ValueError(
            "The ``bond_indices`` set includes atoms not in the ``atom_indices`` set."
        )

    try:
        fragment = _extract_oe_fragment(molecule, atom_indices, bond_indices)
    except (
        ModuleNotFoundError,
        ToolkitUnavailableException,
        LicenseError,
        MissingOptionalDependencyError,
    ):
        fragment = _extract_rd_fragment(molecule, atom_indices, bond_indices)

    # Sanity check that all atoms are still bonded
    fragment_smiles = fragment.to_smiles()

    assert "." not in fragment_smiles, (
        "An atom that is not bonded to any other atom in the fragment. "
        "You probably ran into a bug. Please report the input molecule to the "
        "issue tracker"
    )

    return fragment


def smiles_to_molecule(smiles, add_atom_map: bool = False) -> Molecule:
    """Create a molecule object from an input SMILES pattern.

    Parameters
    ----------
    smiles : str
        SMILES representation of desired molecule.
    add_atom_map
        Whether to create a canonical atom map for the molecule.

    Returns
    -------
        A normalized molecule with desired smiles string.
    """

    molecule = Molecule.from_smiles(smiles, allow_undefined_stereo=True)

    # Add canonical ordered atom maps
    if add_atom_map:
        molecule = molecule.canonical_order_atoms()
        molecule.properties["atom_map"] = {i: i + 1 for i in range(molecule.n_atoms)}

    return molecule



import os
import warnings

import pytest
from openff.fragmenter.depiction import (
    _oe_render_fragment,
    _oe_render_parent,
    _rd_render_fragment,
    _rd_render_parent,
    depict_fragmentation_result,
    depict_fragments,
)
from openff.fragmenter.fragment import Fragment, FragmentationResult
from openff.toolkit.topology import Molecule
from openff.toolkit.utils.exceptions import AtomMappingWarning
from openff.utilities import MissingOptionalDependencyError


@pytest.mark.parametrize("draw_function", [_oe_render_parent, _rd_render_parent])
def test_xx_render_parent(draw_function):
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AtomMappingWarning)

            svg_contents = draw_function(
                Molecule.from_smiles(
                    "[C:1]([H:5])([H:6])([H:7])[C:2]([H:8])([H:9])[C:3]([H:10])([H:11])[C:4]([H:12])([H:13])([H:14])"
                )
            )
    except (ModuleNotFoundError, MissingOptionalDependencyError) as e:
        pytest.skip(str(e))

    assert isinstance(svg_contents, str)
    assert "svg" in svg_contents


@pytest.mark.parametrize("draw_function", [_oe_render_fragment, _rd_render_fragment])
def test_xx_render_fragment(draw_function):
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AtomMappingWarning)
            svg_contents = draw_function(
                Molecule.from_smiles(
                    "[C:1]([H:5])([H:6])([H:7])[C:2]([H:8])([H:9])[C:3]([H:10])([H:11])[C:4]([H:12])([H:13])([H:14])"
                ),
                Molecule.from_smiles(
                    "[C:1]([H:3])([H:4])([H:5])[C:2]([H:6])([H:7])([H:8])"
                ),
                (1, 2),
            )

    except (ModuleNotFoundError, MissingOptionalDependencyError) as e:
        pytest.skip(str(e))

    assert isinstance(svg_contents, str)
    assert "svg" in svg_contents


def test_depict_fragments(tmpdir):
    output_file = os.path.join(tmpdir, "report.html")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AtomMappingWarning)

        depict_fragments(
            Molecule.from_smiles(
                "[C:1]([H:5])([H:6])([H:7])[C:2]([H:8])([H:9])[C:3]([H:10])([H:11])[C:4]([H:12])([H:13])([H:14])"
            ),
            {
                (1, 2): Molecule.from_smiles(
                    "[C:1]([H:3])([H:4])([H:5])[C:2]([H:6])([H:7])([H:8])"
                )
            },
            output_file,
        )

    with open(output_file) as file:
        contents = file.read()

    assert "<html>" in contents
    assert "<img src=" in contents


def test_depict_fragmentation_result(tmpdir):
    output_file = os.path.join(tmpdir, "report.html")

    depict_fragmentation_result(
        FragmentationResult(
            parent_smiles="[C:1]([H:5])([H:6])([H:7])[C:2]([H:8])([H:9])[C:3]([H:10])([H:11])[C:4]([H:12])([H:13])([H:14])",
            fragments=[
                Fragment(
                    smiles="[C:1]([H:3])([H:4])([H:5])[C:2]([H:6])([H:7])([H:8])",
                    bond_indices=(1, 2),
                )
            ],
            provenance={},
        ),
        output_file,
    )

    with open(output_file) as file:
        contents = file.read()

    assert "<html>" in contents
    assert "<img src=" in contents



import json
import os

from openff.toolkit.topology import Molecule
from openff.toolkit.utils import toolkit_registry_manager
from openff.utilities import get_data_file_path


def default_functional_groups() -> dict[str, str]:
    """Returns a dictionary containing the SMARTS representations of the default set
    functional groups which should be preserved during fragmentation (e.g. an amide
    should not be cleaved leaving either a carbonyl or amine group).

    Notes
    -----
    * The groups are loaded from the internal ``data/default-functional-groups.json``
      file.

    Returns
    -------
        A dictionary where each key is the name of a functional group and each value
        the corresponding SMARTS pattern.
    """

    file_name = get_data_file_path(
        os.path.join("data", "default-functional-groups.json"),
        "openff.fragmenter",
    )

    with open(file_name) as file:
        functional_groups = json.load(file)

    return functional_groups


def get_map_index(
    molecule: Molecule, atom_index: int, error_on_missing: bool = True
) -> int:
    """Returns the map index of a particular atom in a molecule.

    Parameters
    ----------
    molecule
        The molecule containing the atom.
    atom_index
        The index of the atom in the molecule.
    error_on_missing
        Whether an error should be raised if the atom does not have a corresponding
        map index

    Returns
    -------
        The map index if found, otherwise 0.
    """
    atom_map = molecule.properties.get("atom_map", {})
    atom_map_index = atom_map.get(atom_index, None)

    if atom_map_index is None and error_on_missing:
        raise KeyError(f"{atom_index} is not in the atom map ({atom_map}).")

    return 0 if atom_map_index is None else atom_map_index


def get_atom_index(molecule: Molecule, map_index: int) -> int:
    """Returns the atom index of the atom in a molecule which has the specified map
    index.

    Parameters
    ----------
    molecule
        The molecule containing the atom.
    map_index
        The map index of the atom in the molecule.

    Returns
    -------
        The corresponding atom index
    """
    inverse_atom_map = {
        j: i for i, j in molecule.properties.get("atom_map", {}).items()
    }

    atom_index = inverse_atom_map.get(map_index, None)

    assert (
        atom_index is not None
    ), f"{map_index} does not correspond to an atom in the molecule."

    return atom_index


# Public with openff-toolkit >=0.14.4
global_toolkit_registry = toolkit_registry_manager



import logging

from openff.toolkit.topology import Molecule

logger = logging.getLogger(__name__)


def _enumerate_stereoisomers(
    molecule: Molecule,
    max_states: int = 200,
    force_flip: bool = True,
) -> list[Molecule]:
    """Enumerate stereoisomers

    Parameters
    ----------
    molecule
        The molecule to enumerate the stereoisomers of.
    max_states
        The max number of states to enumerate
    force_flip
        If True, will flip all stereocenters. If False, will only flip centers that are
        undefined

    Returns
    -------
        The stereoisomers including the input molecule if it had full stereochemistry
        defined.
    """

    if max_states > 200:
        raise NotImplementedError(
            "The max states must currently be less than 200 due to a hard coded maximum "
            "value in the OpenFF toolkit"
        )

    # Check if the input molecule has any undefined stereochemistry. If it does not
    # then it should be included in the list of returned stereoisomers.
    stereoisomers = molecule.enumerate_stereoisomers(
        undefined_only=not force_flip,
        max_isomers=max_states,
        rationalise=False,
    )

    if len(stereoisomers) == 0:
        # Handle the case where the input molecule does not have any stereoisomers.
        return [molecule]

    # Attach the atom map to any stereoisomers.
    if "atom_map" in molecule.properties:
        for stereoisomer in stereoisomers:
            stereoisomer.properties["atom_map"] = molecule.properties["atom_map"]

    # Check to see if the original molecule had undefined stereochemistry.
    undefined_atom_stereochemistry = any(
        atom_old.stereochemistry is None and atom_new.stereochemistry is not None
        for atom_old, atom_new in zip(molecule.atoms, stereoisomers[0].atoms)
    )
    undefined_bond_stereochemistry = any(
        bond_old.stereochemistry is None and bond_new.stereochemistry is not None
        for bond_old, bond_new in zip(molecule.bonds, stereoisomers[0].bonds)
    )

    # If the input molecule had all of the input stereochemistry defined make sure to
    # return it in addition to the newly found stereoisomers.
    if not undefined_atom_stereochemistry and not undefined_bond_stereochemistry:
        stereoisomers = [molecule] + stereoisomers[: max_states - 1]

    return stereoisomers



from contextlib import contextmanager


@contextmanager
def does_not_raise():
    """A helpful context manager to use inplace of a pytest raise statement
    when no exception is expected."""
    yield



"""
openff-fragmenter

Fragment molecules for quantum mechanics torsion scans.
"""

from . import chemi, fragment, utils
from ._version import get_versions

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions

__all__ = [chemi, fragment, utils]



import base64
import os
from collections.abc import Collection

from jinja2 import Template
from openff.fragmenter.fragment import BondTuple, FragmentationResult
from openff.fragmenter.utils import get_map_index
from openff.toolkit.topology import Molecule
from openff.utilities import (
    MissingOptionalDependencyError,
    get_data_file_path,
    requires_oe_module,
)


@requires_oe_module("oechem")
def _oe_fragment_predicates(map_indices: Collection[int]):
    """Returns an atom and bond predicate which matches atoms whose map index
    appears in the specified ``map_indices`` collection"""

    from openeye import oechem

    class PredicateAtoms(oechem.OEUnaryAtomPred):
        def __init__(self, inner_map_indices):
            oechem.OEUnaryAtomPred.__init__(self)
            self.map_indices = inner_map_indices

        def __call__(self, atom):
            return atom.GetMapIdx() in self.map_indices

        def CreateCopy(self):
            return PredicateAtoms(self.map_indices).__disown__()

    class PredicateBonds(oechem.OEUnaryBondPred):
        def __init__(self, inner_map_indices):
            oechem.OEUnaryBondPred.__init__(self)
            self.map_indices = inner_map_indices

        def __call__(self, bond: oechem.OEBondBase):
            return (
                bond.GetBgn().GetMapIdx() in self.map_indices
                and bond.GetEnd().GetMapIdx() in self.map_indices
            )

        def CreateCopy(self):
            return PredicateBonds(self.map_indices).__disown__()

    return PredicateAtoms(map_indices), PredicateBonds(map_indices)


@requires_oe_module("oedepict")
def _oe_wbo_label_display(bond_tuples: Collection[BondTuple]):
    """Returns a ``OEDisplayBondPropBase`` subclass which will label bonds with the
    specified map indices with their WBO values if present.
    """

    from openeye import oedepict

    class LabelWibergBondOrder(oedepict.OEDisplayBondPropBase):
        def __init__(self, inner_bond_tuples):
            oedepict.OEDisplayBondPropBase.__init__(self)
            self.bond_tuples = inner_bond_tuples

        def __call__(self, bond):
            map_tuple = (bond.GetBgn().GetMapIdx(), bond.GetEnd().GetMapIdx())

            if (
                map_tuple not in self.bond_tuples
                and tuple(reversed(map_tuple)) not in self.bond_tuples
            ):
                return " "

            if "fractional_bond_order" not in bond.GetData():
                return " "

            return f"{bond.GetData('fractional_bond_order'):.2f}"

        def CreateCopy(self):
            return LabelWibergBondOrder(self.bond_tuples).__disown__()

    return LabelWibergBondOrder(bond_tuples)


@requires_oe_module("oedepict")
def _oe_render_parent(
    parent: Molecule,
    rotor_bonds: Collection[BondTuple] | None = None,
    image_width: int = 572,
    image_height: int = 198,
) -> str:
    from openeye import oedepict

    rotor_bonds = [] if rotor_bonds is None else rotor_bonds

    # Map the OpenFF molecules into OE ones, making sure to explicitly set the atom
    # map on the OE object as this is not handled by the OpenFF toolkit.
    oe_parent = parent.to_openeye()

    for atom in oe_parent.GetAtoms():
        atom.SetMapIdx(get_map_index(parent, atom.GetIdx(), False))

    oedepict.OEPrepareDepiction(oe_parent)

    # Set-up common display options.
    image = oedepict.OEImage(image_width, image_height)

    display_options = oedepict.OE2DMolDisplayOptions(
        image_width, image_height, oedepict.OEScale_AutoScale
    )
    display_options.SetTitleLocation(oedepict.OETitleLocation_Hidden)
    display_options.SetAtomColorStyle(oedepict.OEAtomColorStyle_WhiteMonochrome)
    display_options.SetAtomLabelFontScale(1.2)
    display_options.SetBondPropertyFunctor(_oe_wbo_label_display(rotor_bonds))

    display = oedepict.OE2DMolDisplay(oe_parent, display_options)

    oedepict.OERenderMolecule(image, display)

    svg_contents = oedepict.OEWriteImageToString("svg", image)
    return svg_contents.decode()


@requires_oe_module("oedepict")
@requires_oe_module("oechem")
def _oe_render_fragment(
    parent: Molecule,
    fragment: Molecule,
    bond_indices: BondTuple,
    image_width: int = 283,
    image_height: int = 169,
) -> str:
    from openeye import oechem, oedepict

    # Map the OpenFF molecules into OE ones, making sure to explicitly set the atom
    # map on the OE object as this is not handled by the OpenFF toolkit.
    oe_parent = parent.to_openeye()

    for atom in oe_parent.GetAtoms():
        atom.SetMapIdx(get_map_index(parent, atom.GetIdx(), False))

    oedepict.OEPrepareDepiction(oe_parent)

    oe_fragment = fragment.to_openeye()

    for atom in oe_fragment.GetAtoms():
        atom.SetMapIdx(get_map_index(fragment, atom.GetIdx(), False))

    oe_parent_bond = oe_parent.GetBond(
        oe_parent.GetAtom(oechem.OEHasMapIdx(bond_indices[0])),
        oe_parent.GetAtom(oechem.OEHasMapIdx(bond_indices[1])),
    )

    # Set-up common display options.
    image = oedepict.OEImage(image_width, image_height)

    display_options = oedepict.OE2DMolDisplayOptions(
        image_width, image_height, oedepict.OEScale_AutoScale
    )

    display_options.SetTitleLocation(oedepict.OETitleLocation_Hidden)
    display_options.SetAtomColorStyle(oedepict.OEAtomColorStyle_WhiteMonochrome)
    display_options.SetAtomLabelFontScale(1.2)

    # display_options.SetBondPropertyFunctor(_oe_wbo_label_display({bond_indices}))

    display = oedepict.OE2DMolDisplay(oe_parent, display_options)

    fragment_atom_predicate, fragment_bond_predicate = _oe_fragment_predicates(
        {atom.GetMapIdx() for atom in oe_fragment.GetAtoms()}
    )

    not_fragment_atoms = oechem.OENotAtom(fragment_atom_predicate)
    not_fragment_bonds = oechem.OENotBond(fragment_bond_predicate)

    oedepict.OEAddHighlighting(
        display,
        oedepict.OEHighlightByColor(oechem.OEGrey, 0.75),
        not_fragment_atoms,
        not_fragment_bonds,
    )

    rotatable_bond = oechem.OEAtomBondSet()

    rotatable_bond.AddBond(oe_parent_bond)
    rotatable_bond.AddAtom(oe_parent_bond.GetBgn())
    rotatable_bond.AddAtom(oe_parent_bond.GetEnd())

    oedepict.OEAddHighlighting(
        display,
        oechem.OEColor(oechem.OELimeGreen),
        oedepict.OEHighlightStyle_BallAndStick,
        rotatable_bond,
    )

    oedepict.OERenderMolecule(image, display)

    svg_contents = oedepict.OEWriteImageToString("svg", image)
    return svg_contents.decode()


def _rd_render_parent(
    parent: Molecule,
    image_width: int = 572,
    image_height: int = 198,
) -> str:
    from rdkit import Chem
    from rdkit.Chem import Draw
    from rdkit.Chem.rdDepictor import Compute2DCoords

    rd_parent: Chem.Mol = parent.to_rdkit()
    rd_parent = Chem.RemoveHs(rd_parent)
    Compute2DCoords(rd_parent)

    drawer = Draw.MolDraw2DSVG(image_width, image_height)
    drawer.DrawMolecule(rd_parent)
    drawer.FinishDrawing()

    svg_contents = drawer.GetDrawingText()

    return svg_contents


def _rd_render_fragment(
    parent: Molecule,
    fragment: Molecule,
    bond_indices: BondTuple,
    image_width: int = 283,
    image_height: int = 169,
) -> str:
    from rdkit import Chem
    from rdkit.Chem import Draw
    from rdkit.Chem.rdDepictor import Compute2DCoords

    rd_parent: Chem.Mol = parent.to_rdkit()

    for atom in rd_parent.GetAtoms():
        atom.SetAtomMapNum(get_map_index(parent, atom.GetIdx(), False))

    rd_parent = Chem.RemoveHs(rd_parent)
    Compute2DCoords(rd_parent)

    map_indices = {*fragment.properties["atom_map"].values()} - {0}

    fragment_atom_indices = [
        atom.GetIdx()
        for atom in rd_parent.GetAtoms()
        if atom.GetAtomMapNum() in map_indices
    ]
    fragment_bond_indices = [
        bond.GetIdx()
        for bond in rd_parent.GetBonds()
        if bond.GetBeginAtom().GetAtomMapNum() in map_indices
        and bond.GetEndAtom().GetAtomMapNum() in map_indices
    ]

    rotatable_bond_index = [
        bond.GetIdx()
        for bond in rd_parent.GetBonds()
        if bond.GetBeginAtom().GetAtomMapNum() in bond_indices
        and bond.GetEndAtom().GetAtomMapNum() in bond_indices
    ]

    for atom in rd_parent.GetAtoms():
        atom.SetAtomMapNum(0)

    drawer = Draw.MolDraw2DSVG(image_width, image_height)

    draw_options = drawer.drawOptions()
    draw_options.useBWAtomPalette()

    drawer.DrawMolecule(
        rd_parent,
        highlightAtoms=fragment_atom_indices,
        highlightAtomColors={
            index: (52.0 / 255.0, 143.0 / 255.0, 235.0 / 255.0)
            for index in fragment_atom_indices
        },
        highlightBonds=fragment_bond_indices + rotatable_bond_index,
        highlightBondColors={
            index: (
                (239.0 / 255.0, 134.0 / 255.0, 131.0 / 255.0)
                if index in rotatable_bond_index
                else (52.0 / 255.0, 143.0 / 255.0, 235.0 / 255.0)
            )
            for index in fragment_bond_indices + rotatable_bond_index
        },
    )
    drawer.FinishDrawing()

    svg_contents = drawer.GetDrawingText()

    return svg_contents


def _compress_svg(svg_contents: str) -> str:
    encoded_image = base64.b64encode(svg_contents.encode()).decode()
    return f"data:image/svg+xml;base64,{encoded_image}"


def depict_fragmentation_result(result: FragmentationResult, output_file: str):
    """Generates a HTML report of fragments for a parent molecule with the rotatable
    bond highlighted.

    Parameters
    ----------
    result
        The result of fragmenting a molecule with a fragmentation engine.
    output_file : str
        The name of the file to write out to.
    """

    depict_fragments(
        result.parent_molecule,
        {fragment.bond_indices: fragment.molecule for fragment in result.fragments},
        output_file,
    )


def depict_fragments(
    parent: Molecule, fragments: dict[BondTuple, Molecule], output_file: str
):
    """Generates a HTML report of fragments for a parent molecule with the rotatable
    bond highlighted.

    Parameters
    ----------
    parent
        The parent molecule that was fragmented.
    fragments
        The fragments generated around each rotatable bond.
    output_file : str
        The name of the file to write out to.
    """

    try:
        header_svg = _oe_render_parent(parent, [*fragments])
        fragment_svg = [
            _oe_render_fragment(parent, fragment, bond_tuple)
            for bond_tuple, fragment in fragments.items()
        ]

    except (ModuleNotFoundError, MissingOptionalDependencyError):
        header_svg = _rd_render_parent(parent)
        fragment_svg = [
            _rd_render_fragment(parent, fragment, bond_tuple)
            for bond_tuple, fragment in fragments.items()
        ]

    template_path = get_data_file_path(
        os.path.join("data", "report-template.html"),
        "openff.fragmenter",
    )

    with open(template_path) as file:
        template = Template(file.read())

    rendered = template.render(
        parent=_compress_svg(header_svg),
        fragments=[_compress_svg(svg) for svg in fragment_svg],
    )

    with open(output_file, "w") as file:
        file.write(rendered)



"""Test chemi module"""

import warnings

import numpy
import pytest
from openff.fragmenter import chemi
from openff.fragmenter._tests.utils import using_openeye
from openff.fragmenter.chemi import (
    _extract_oe_fragment,
    _extract_rd_fragment,
    _find_oe_stereocenters,
    _find_rd_stereocenters,
    assign_elf10_am1_bond_orders,
    extract_fragment,
    find_ring_systems,
    find_stereocenters,
    smiles_to_molecule,
)
from openff.fragmenter.utils import global_toolkit_registry
from openff.toolkit.topology import Molecule
from openff.toolkit.utils import (
    AmberToolsToolkitWrapper,
    OpenEyeToolkitWrapper,
    ToolkitRegistry,
)
from openff.toolkit.utils.exceptions import AtomMappingWarning
from openff.units import unit
from openff.utilities import MissingOptionalDependencyError


def test_assign_elf10_am1_bond_orders():
    molecule = assign_elf10_am1_bond_orders(Molecule.from_smiles("CCCC"))

    for bond in molecule.bonds:
        assert bond.fractional_bond_order is not None


@using_openeye
def test_assign_elf10_am1_bond_orders_simple_parity():
    with global_toolkit_registry(OpenEyeToolkitWrapper()):
        molecule = assign_elf10_am1_bond_orders(Molecule.from_smiles("C"))
        oe_bond_orders = [bond.fractional_bond_order for bond in molecule.bonds]

    with global_toolkit_registry(
        ToolkitRegistry([AmberToolsToolkitWrapper(), OpenEyeToolkitWrapper()])
    ):
        molecule = assign_elf10_am1_bond_orders(Molecule.from_smiles("C"))
        at_bond_orders = [bond.fractional_bond_order for bond in molecule.bonds]

    assert numpy.allclose(oe_bond_orders, at_bond_orders, atol=1.0e-2)


@pytest.mark.parametrize("smiles, max_confs", [("CCCCCCC", 1), ("CCCCCCC", 3)])
def test_generate_conformers(smiles, max_confs):
    returned_molecule = chemi._generate_conformers(
        Molecule.from_smiles(smiles), max_confs=max_confs
    )

    assert 0 < returned_molecule.n_conformers <= max_confs


def test_generate_conformers_ordering():
    original_molecule = Molecule.from_smiles("CCCC")

    returned_molecule = chemi._generate_conformers(original_molecule, max_confs=1)
    assert returned_molecule.n_conformers == 1

    # Make sure the atom ordering did not change.
    _, atom_map = Molecule.are_isomorphic(
        original_molecule, returned_molecule, return_atom_map=True
    )

    assert all(i == j for i, j in atom_map.items())


def test_generate_conformers_canonical_check():
    original_molecule = Molecule.from_smiles("CCCCl").canonical_order_atoms()
    original_molecule = chemi._generate_conformers(original_molecule, max_confs=1)

    # Generate a conformer using a molecule with permuted atom orderings.
    atom_map = numpy.arange(original_molecule.n_atoms)
    numpy.random.shuffle(atom_map)

    remapped_molecule = original_molecule.remap(
        {int(i): int(j) for i, j in enumerate(atom_map)}
    )
    remapped_molecule = chemi._generate_conformers(remapped_molecule, max_confs=1)

    original_conformer = original_molecule.conformers[0].m_as(unit.angstrom)
    remapped_conformer = remapped_molecule.conformers[0].m_as(unit.angstrom)

    assert numpy.allclose(original_conformer[:4], remapped_conformer[atom_map][:4])


@using_openeye
@pytest.mark.parametrize(
    "smiles",
    [
        "c1ccccc1",
        "C1CC2CCC1C2",
        "C12C3C4C1C5C2C3C45",
        "c1ccc2ccccc2c1",
        "c1ccc(cc1)C2CCCC2",
        "c1ccc(cc1)C2CCC3C2CC3",
        "c1cc2ccc3cccc4c3c2c(c1)cc4",
        "C1CCC2(CC1)CCCCC2",
        "c1ccc(cc1)Cc2ccccc2",
    ],
)
def test_find_ring_systems(smiles):
    from openeye import oechem

    molecule = Molecule.from_smiles(smiles, allow_undefined_stereo=True)

    oe_molecule = molecule.to_openeye()
    _, expected = oechem.OEDetermineRingSystems(oe_molecule)

    ring_systems = find_ring_systems(molecule)

    for i, ring_system in enumerate(expected):
        if ring_system > 0:
            assert ring_systems[i] == ring_system

        else:
            assert i not in ring_systems


@pytest.mark.parametrize("add_atom_map", [True, False])
def test_smiles_to_molecule(add_atom_map):
    molecule = smiles_to_molecule("CCCC", add_atom_map=add_atom_map)

    assert isinstance(molecule, Molecule)
    assert ("atom_map" in molecule.properties) == add_atom_map


@pytest.mark.parametrize(
    "smiles, expected_atoms, expected_bonds",
    [
        ("[C:1]([H:6])([H:7])([H:8])[C:5]([Br:3])([Cl:4])([H:2])", [4], []),
        ("[C:3]([Cl:1])([H:2])=[C:6]([H:4])([Cl:5])", [], [(2, 5)]),
        (
            "[C:3]([Cl:1])([H:2])=[C:6]([H:4])[C:5]([Br:7])([Cl:8])([H:9])",
            [4],
            [(2, 5)],
        ),
    ],
)
@pytest.mark.parametrize(
    "find_method", [_find_oe_stereocenters, _find_rd_stereocenters, find_stereocenters]
)
def test_find_stereocenters(smiles, expected_atoms, expected_bonds, find_method):
    molecule = Molecule.from_mapped_smiles(smiles, allow_undefined_stereo=True)

    stereogenic_atoms, stereogenic_bonds = None, None

    try:
        stereogenic_atoms, stereogenic_bonds = find_method(molecule)
    except (ModuleNotFoundError, MissingOptionalDependencyError) as e:
        pytest.skip(str(e))

    assert stereogenic_atoms == expected_atoms
    assert stereogenic_bonds == expected_bonds


@pytest.mark.parametrize(
    "smiles, atoms, bonds, expected",
    [
        (
            "[C:1]([H:4])([H:5])([H:6])"
            "[C:2]([H:7])([H:8])"
            "[C:3]([H:9])([H:10])([H:11])",
            {1, 2},
            {(1, 2)},
            "CC",
        ),
        (
            "[C:1]([H:4])([H:5])([H:6])"
            "[C:2]([H:7])([H:8])"
            "[C:3]([H:9])([H:10])([H:11])",
            {1, 2, 4},
            {(1, 2), (1, 4)},
            "CC",
        ),
        (
            r"[H:6]/[C:1](=[C:2](\[C:3]([H:7])([H:8])[H:9])/[Cl:5])/[Cl:4]",
            {1, 2, 4, 5},
            {(1, 2), (1, 4), (2, 5)},
            r"Cl\C=C/Cl",
        ),
        (
            "[H:7][C:1]([H:8])([C@:2]([F:3])([Cl:5])[Br:6])[Cl:4]",
            {1, 2, 3, 5, 6},
            {(1, 2), (2, 3), (2, 5), (2, 6)},
            r"C[C@](F)(Cl)Br",
        ),
    ],
)
@pytest.mark.parametrize(
    "extract_method", [_extract_rd_fragment, _extract_oe_fragment, extract_fragment]
)
def test_extract_fragment(smiles, atoms, bonds, expected, extract_method):
    molecule = Molecule.from_mapped_smiles(smiles)
    molecule.properties["atom_map"] = {i: i + 1 for i in range(molecule.n_atoms)}

    fragment = None

    try:
        fragment = extract_method(molecule, atoms, bonds)
    except (ModuleNotFoundError, MissingOptionalDependencyError) as e:
        pytest.skip(str(e))

    expected_fragment = Molecule.from_smiles(expected)

    assert Molecule.are_isomorphic(
        fragment, expected_fragment, bond_stereochemistry_matching=False
    )[0]


def test_extract_fragment_bonds_in_atoms():
    """Tests that an exception is raised when the bonds set contains atoms not in the
    atoms set."""

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AtomMappingWarning)

        molecule = Molecule.from_smiles("[H:1][C:2]#[C:3][H:4]")

    with pytest.raises(ValueError, match="set includes atoms not in the"):
        extract_fragment(molecule, {2}, {(2, 3)})


@using_openeye
def test_extract_fragment_disconnected_fragment_warning():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AtomMappingWarning)

        molecule = Molecule.from_smiles(
            "[C:1]([H:3])([H:4])([H:5])[C:2]([H:6])([H:7])([H:8])"
        )

    with pytest.raises(AssertionError, match="An atom that is not bonded"):
        extract_fragment(molecule, {1, 2}, set())



"""
openff-fragmenter

Fragment molecules for quantum mechanics torsion scans.
"""

import sys

from setuptools import find_namespace_packages, setup

import versioneer

short_description = __doc__.split("\n")

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []

try:
    with open("README.md") as handle:
        long_description = handle.read()
except OSError:
    long_description = "\n".join(short_description[2:])


setup(
    name="acellera-openff-fragmenter",
    author="Chaya D. Stern",
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="MIT",
    packages=find_namespace_packages(include=["openff.*"]),
    include_package_data=True,
    setup_requires=[] + pytest_runner,
    python_requires=">=3.6",
)



import importlib
from collections import defaultdict
from typing import TypeVar

import pytest
from openff.fragmenter.fragment import BondTuple
from openff.fragmenter.utils import get_map_index
from openff.toolkit.topology import Molecule

T = TypeVar("T")

try:
    importlib.import_module("openeye.oechem")

    has_openeye = True
except ImportError:
    has_openeye = False

using_openeye = pytest.mark.skipif(not has_openeye, reason="Cannot run without OpenEye")


def smarts_set_to_map_indices(
    input_set: set[str], molecule: Molecule
) -> set[int | BondTuple]:
    """A utility function which maps a set of the form ``{"SMARTS_1", "SMARTS_2", ...}``
    to one of the form ``{map_index_1, map_index_2, ...}`` when the SMARTS defines a
    single atom or ``{(map_index_1_a, map_index_1_b), ...}`` when the SMARTS defines a
    bond.

    The is useful for tests which need to define expected values for atoms / bonds
    with specific map indices in a way that is agnostic of how the molecule used in
    the test is ordered (i.e. whether it was canonically ordered with RDKit or OE.
    """

    return_value = set()

    for smarts in input_set:
        matches = list(
            {
                tuple(sorted(match))
                for match in molecule.chemical_environment_matches(smarts)
            }
        )
        assert len(matches) == 1

        match = matches[0]
        assert 0 < len(match) <= 2

        if len(match) == 1:
            return_value.add(get_map_index(molecule, match[0]))
        else:
            return_value.add(tuple(get_map_index(molecule, i) for i in match))

    return return_value


def key_smarts_to_map_indices(
    input_dictionary: dict[str, T], molecule: Molecule
) -> dict[int | BondTuple, T]:
    """A utility function which maps a dictionary of the form ``value["SMARTS"] = x``
    to one of the form ``value[map_index] = x`` when the SMARTS defines a single atom
    or ``value[(map_index_1, map_index_2)] = x`` when the SMARTS defines a bond.

    The is useful for tests which need to define expected values for atoms / bonds
    with specific map indices in a way that is agnostic of how the molecule used in
    the test is ordered (i.e. whether it was canonically ordered with RDKit or OE.
    """

    return_value = {}

    for smarts, expected in input_dictionary.items():
        matches = list(
            {
                tuple(sorted(match))
                for match in molecule.chemical_environment_matches(smarts)
            }
        )
        assert len(matches) == 1

        match = matches[0]
        assert 0 < len(match) <= 2

        if len(match) == 1:
            return_value[get_map_index(molecule, match[0])] = expected
        else:
            return_value[tuple(get_map_index(molecule, i) for i in match)] = expected

    return return_value


def value_smarts_to_map_indices(
    input_dictionary: dict[str, set[str]], molecule: Molecule
) -> dict[str, set[int | BondTuple]]:
    """A utility function which maps a dictionary of the form
    ``value[key] = {"SMARTS_1", "SMARTS_2", ...}`` to one of the form
    ``value[key] = {map_index_1, map_index_2, ...}`` when the SMARTS defines a single
    atom or ``value[key] = {(map_index_1_a, map_index_1_b), ...}`` when the SMARTS
    defines a bond.

    The is useful for tests which need to define expected values for atoms / bonds
    with specific map indices in a way that is agnostic of how the molecule used in
    the test is ordered (i.e. whether it was canonically ordered with RDKit or OE.
    """

    return_value = defaultdict(set)

    for key, values in input_dictionary.items():
        for smarts in values:
            matches = list(
                {
                    tuple(sorted(match))
                    for match in molecule.chemical_environment_matches(smarts)
                }
            )

            for match in matches:
                assert 0 < len(match) <= 2

                if len(match) == 1:
                    return_value[key].add(get_map_index(molecule, match[0]))
                else:
                    return_value[key].add(
                        tuple(get_map_index(molecule, i) for i in match)
                    )

    return {**return_value}
